module(..., package.seeall)
local s = require'scheme'

local clonePath = s.clonePath
local function getParent(path)
  local path = clonePath(path, -1)
  local sch = C.getSchemeByPath(path)
  local conf = C.getConfByPath(path) or {}
  conf = s.addDefault(path, sch, conf, true) --clone conf with defaults
  return conf
end

local setupOptions = {
  protostack = "protostack",
  ["nat-traversal"] = "nat_traversal",
}

local connOptions = {
  left = "left",
  leftid = "leftid",
  leftnexthop = "leftnexthop",
  leftsubnet = "leftsubnet",
  leftrsasigkey = "leftrsasigkey",
  leftsourceip = "leftsourceip",
  leftcert = "leftcert",

  right = "right",
  rightid = "rightid",
  rightnexthop = "rightnexthop",
  rightsubnet = "rightsubnet",
  rightrsasigkey = "rightrsasigkey",
  rightsourceip = "rightsourceip",
  rightcert = "rightcert",

  auto = "auto",
  authby = "authby",
  keylife = "keylife",
  ikelifetime = "ikelifetime",
  pfs = "pfs",
  dpddelay = "dpddelay",
  dpdtimeout = "dpdtimeout",
}

local optionTransform = {
  keylife = function (val)
    return val..'s'
  end,
  ikelifetime = function (val)
    return val..'s'
  end,
}

local function createIpsecConfFile(conf)
  local ipsecConf = {
    (BI.arch == 'nsg700' and '#' or '').."version 2.0", --TODO
    "config setup",
  }
  local setup = conf.setup or {}

  for k, v in pairs(setup) do
    if setupOptions[k] then
      ipsecConf[#ipsecConf+1] = 
      ("\t%s=%s"):format(setupOptions[k], tostring(v))
    end
  end
  for k, v in pairs(setup['extra-options'] or {}) do
    ipsecConf[#ipsecConf+1] = 
    ("\t%s=%s"):format(k, tostring(v))
  end

  local lra = {}
  for con, conConf in pairs(conf.connections or {}) do
    ipsecConf[#ipsecConf+1] = "conn "..con
    for k, v in pairs(conConf or {}) do
      if connOptions[k] then
        if optionTransform[k] then
          v = optionTransform[k](v)
        end
        ipsecConf[#ipsecConf+1] = 
        ("\t%s=%s"):format(connOptions[k], tostring(v))
      end
    end

    local esp = {}
    for k, v in pairs(conConf.esp or {}) do
      if k == "extra" then
        if not v:match"^%s*$" then
          esp[#esp+1] = v
        end
      else
        esp[#esp+1] = k
      end
    end
    if #esp > 0 then
      ipsecConf[#ipsecConf+1] = ("\tesp=%s"):format(table.concat(esp, ','))
    end

    for k, v in pairs(conConf['extra-options'] or {}) do
      ipsecConf[#ipsecConf+1] = 
      ("\t%s=%s"):format(k, tostring(v))
    end
    -- add left/right adresses
    if conConf.left then
       lra[conConf.left] = true
    end
    if conConf.right then
       lra[conConf.right] = true
    end
  end
  local ipsecConfText = table.concat(ipsecConf, '\n')..'\n'
  local f = io.open("/etc/ipsec.conf", 'w')
  if f then
    f:write(ipsecConfText)
    f:close()
  end
  return lra
end

local function createIpsecSecretsFile(conf)
  local ipsecSecrets = {}

  for _, rsa in ipairs(conf.rsa or {}) do
    local secret = rsa.secret
    local file = rsa.file
  
     if file and not file:match"^%s*$" then
	ipsecSecrets[#ipsecSecrets+1] = table.concat(rsa.indices or {}, ' ')..
	   ": RSA\t"..file.."\t"..(secret or "")  -- secret may be used as a password to "key"-file
     else
	if secret and not secret:match"^%s*$" then
	   ipsecSecrets[#ipsecSecrets+1] = table.concat(rsa.indices or {}, ' ')..
	      ": RSA\t{\n"..secret.."\t}"
	end
     end
  end

  for _, psk in ipairs(conf.psk or {}) do
    local secret = psk.secret
    if secret and not secret:match"^%s*$" then
      ipsecSecrets[#ipsecSecrets+1] = table.concat(psk.indices or {}, ' ')..
      ': PSK\t"'..secret..'"'
    end
  end

  local ipsecSecretsText = table.concat(ipsecSecrets, '\n')..'\n'
  local f = io.open("/etc/ipsec.secrets", 'w')
  if f then
    f:write(ipsecSecretsText)
    f:close()
  end
end

local ipsecNeteventKey = {}

local function ipsecAtStart(oper)
  local fib = 1
  local tmo = 2
  while true do
    local dat = os.date'*t'
    if oper.daemon == 'kill' or (dat.year >= 2010 and dat.month >= 2) then
      break
    end
    oper.log:write(os.date('%F %T').." Real time seems not set. Wait "..
      tmo.." sec.\n"
    )
    C.threadSleep(tmo, oper)
    if tmo < 300 then
      local prevTmo = tmo
      tmo = tmo + fib
      fib = prevTmo
    end
  end
end

local function testTime(conf)
  for con, conConf in pairs(conf.connections or {}) do
    if conConf.leftrsasigkey == "%cert" or conConf.rightrsasigkey == "%cert" then
      return true
    end
  end
  return false
end

local function restartIpsec(path, sch, conf, prevConf)
  if not s.confCmp(conf, prevConf) then  
    if conf and conf.enable then
      local conf = s.addDefault(path, sch, conf, true) --clone conf with defaults
      local lra = createIpsecConfFile(conf)
      createIpsecSecretsFile(conf.secrets or {})
      local locAtStart = testTime(conf) and ipsecAtStart
      local ipsec_restart = function()
        C.restartDaemon(
          path, 
          {"/etc/init.d/ipsec", "restart"}, 
          { noRestart = true, usePopen = true,
            atStart = locAtStart,
          }
        )
      end
      C.updateNeteventCallback(
        ipsecNeteventKey,
        function(msg)
          if ((msg.type == 'newaddr' and lra[msg.address])  or
            (msg.type == 'newroute' and lra['%defaultroute'] and msg.netlen == 0)
          ) then
            ipsec_restart()
          end
        end

      )
      ipsec_restart()
    else
      C.restartDaemon(
        path, 
        {"/etc/init.d/ipsec", "stop"}, 
        {noRestart = true, usePopen = true}
      ) 
    end
  end
  return true
end

local indicesScheme = s.newScheme{
  _H = "ipsecSecretsIndices",
  _LIST = s.stringType{
    _H = "ipsecSecretsIndex",
  },
}

local secretsScheme = s.newScheme{
  _H = "ipsecSecrets",
  psk = s.newScheme{
    _H = "ipsecSecretsPsklist",
    _LIST = s.newScheme{
      _H = "ipsecSecretsPsk",
      indices = indicesScheme,
      secret = s.stringType{
        _H = "ipsecSecretsPskSecret",
      }
    }
  },
  rsa = s.newScheme{
    _H = "ipsecSecretsRsalist",
    _LIST = s.newScheme{
      _H = "ipsecSecretsRsa",
       indices = indicesScheme{_P = 30, },
      file = s.stringType{
	_P = 25,
        _H = "ipsecSecretsRsaFile",
      }, 
      secret = s.stringType{
	_P = 24,
        _H = "ipsecSecretsRsaSecret",
      },
      generate = s.newScheme{
        _P = 14,
        _H = "ipsecSecretsRsaGenerate",
        _T = "string",
        _EX = { subtype = 'areyousure'},
        _V = s.areyousureVal,
        _A = {"\nNew RSA key will be generated.\nIt takes a lot of time.\n"},
        _X = function(path, sch, val, sesNum)
          local com
          local yes = val == 'yes'
          if yes then
            com = {
              '/usr/local/sbin/ipsec', 'rsasigkey', 
              '--random', '/dev/urandom', 
              getParent(path).keylengh or 2192
            }
          else
            com = {'echo', 'RSA key NOT generated.'}
          end
          local myLog = ''
          C.restartDaemon(path, com,
            {sesNum = sesNum, noTimestamp = true, noRestart = true,
              atRead = yes and function(oper)
                myLog = myLog..oper.log:read(-1)
              end,
              atQuit = yes and function(oper)
                C.setConfigValue(clonePath(path, -1, "secret"), myLog)
              end,
          }) 
        end,
        _XI = true,
      },
      keylengh = s.newScheme{
       _P = 15,
       _H = "ipsecSecretsRsaKeylengh",
        _T = "number",
        _A = {16, 4096},
        _V = s.integerVal,
        _D = 2192,
      },
      ["show-public-key"] = s.newScheme{
        _P = 5,
        _H = "ipsecSecretsShowpublickey",
        _X = function(path)
          local secret = C.getConfByPath(clonePath(path, -1, 'secret'))
          if secret and not secret:match"^%s*$" then
            local tmpSecretFileName = os.tmpname()
            local f = io.open(tmpSecretFileName, 'w')
            f:write(": RSA\t{\n"..secret.."\t}\n")
            f:close()
            local out = ('\n'..s.readpopen("/usr/local/sbin/ipsec showhostkey --left --file "..
              tmpSecretFileName.." 2>&1"):
              gsub('^.*leftrsasigkey=(.*)$', '%1')
            )
            os.remove(tmpSecretFileName)
            return out
          end
          return "No RSA secret"
        end,
      },

    },
  },
  --[[
  ["show-hostkey"] = s.newScheme{
    _H = helpX.ipsecSecretsShowHostkey,
    _X = function(path)
      createIpsecSecretsFile(C.getConfByPath(clonePath(path, -1)) or {})
      return('\n'..s.readpopen"ipsec showhostkey --left 2>&1":
        gsub('^.*leftrsasigkey=(.*)$', '%1')
      )
    end,
  },
  --]]
  --[[
  ["show-secrets-file"] = s.newScheme{
    _X = function()
      return('\n'..s.readpopen("cat /etc/ipsec.secrets 2>&1"))
    end,
  },
  --]]
}

local setupScheme = s.newScheme{
  _H = "ipsecSetup",

  protostack = s.newScheme{
    _H = "ipsecSetupProtostack",
    _T = 'choice',
    _A = {
      {"netkey"},
      --[[ -Dan 5.10.2010 The all values except for "netkey" are deleted (may be temporary)
      {"klips"},
      {"mast"},
      {"auto"},
      --]]
    },
    _V = s.choiceVal,
    _D = "netkey",
  },

  ["nat-traversal"] = s.newScheme{
    _H = "ipsecSetupNatTraversal",
    _T = 'choice',
    _A = {"yes", "no"},
    _V = s.choiceVal,
    _D = "yes",
  },
  ["extra-options"] = s.newScheme{
    _H = "ipsecSetupExtraoptions",
    _P = -100,
    _NAME = s.stringType{
      _H = "ipsecSetupExtraoptionsOption",
    }
  },
}

local leftRightValues = {
  "%defaultroute",
  "%any",
  "%opportunistic",
}

local leftRightIdValues = {
  "%myid",
}

local leftRightNexthopValues = {
  "%defaultroute",
  "%direct",
}

local function getUnicKey(str, list)
  local res
  for _, pattern in ipairs(list) do
    if str == pattern:sub(1, #str) then
      if res then
        return nil
      end
      res = pattern
    end
  end
  return res
end

local function ipaddressOrKeyVal(val, keytables)
  local res = getUnicKey(val, keytables)
  if res or s.stringVal(val) then
    return res or val
  end
  return nil, '\n'..table.concat(leftRightValues, ',\n')..'\nIP address or hostname'
end


local leftRightScheme = s.newScheme{
  _T = "string",
  _V = ipaddressOrKeyVal,
  _A = {leftRightValues},
}

local leftRightIdScheme = s.newScheme{
  _T = "string",
  _V = ipaddressOrKeyVal,
  _A = {leftRightIdValues},
}

local leftRightNexthopScheme = s.newScheme{
  _T = "string",
  _V = ipaddressOrKeyVal,
  _A = {leftRightNexthopValues},
}

local ipsecConnectionScheme = s.newScheme{
  _H = "ipsecTun",
  left = leftRightScheme{_H = "ipsecTunLeft"},
  leftid = leftRightIdScheme{_H = "ipsecTunLeftid"},
  leftnexthop = leftRightNexthopScheme{_H = "ipsecTunLeftnexthop"},
  leftsubnet = s.ipPrefixType{_H = "ipsecTunLeftsubnet"},
  leftrsasigkey = s.stringType{_H = "ipsecTunLeftrsasigkey"},
  leftsourceip = s.ipAddressType{_H = "ipsecTunLeftsourceip"},
  leftcert = s.stringType{_H = "ipsecTunLeftcert"},

  right = leftRightScheme{_H = "ipsecTunRight"},
  rightid = leftRightIdScheme{_H = "ipsecTunRightid"},
  rightnexthop = leftRightNexthopScheme{_H = "ipsecTunRightnexthop"},
  rightsubnet = s.ipPrefixType{_H = "ipsecTunRightsubnet"},
  rightrsasigkey = s.stringType{_H = "ipsecTunRightrsasigkey"},
  rightsourceip = s.ipAddressType{_H = "ipsecTunRightsourceip"},
  rightcert = s.stringType{_H = "ipsecTunRightcert"},

  auto = s.newScheme{
    _H = "ipsecTunAuto",
    _T = 'choice',
    _A = {
      {"add", ""},
      {"route", ""},
      {"start", ""},
      {"manual", ""},
      {"ignore", ""},
    },
    _V = s.choiceVal,
    _D = "ignore",
  },

  authby = s.newScheme{
    _H = "ipsecTunAuthby",
    _T = 'choice',
    _A = {
      {"rsasig", ""},
      {"secret", ""},
      {"secret|rsasig", ""},
      {"never", ""},
    },
    _V = s.choiceVal,
    _D = "rsasig",
  },

  esp = s.newScheme{
    _H = "ipsecTunEsp",
    ["aes"] = s.booleanType{_H = "ipsecTunEspAes"},
    ["aes-md5"] = s.booleanType{_H = "ipsecTunEspAesMd5"},
    ["aes-sha1"] = s.booleanType{_H = "ipsecTunEspAesSha1"},
    ["3des"] = s.booleanType{_H = "ipsecTunEsp3des"},
    ["3des-md5"] = s.booleanType{_H = "ipsecTunEsp3desMd5"},
    ["3des-sha1"] = s.booleanType{_H = "ipsecTunEsp3desSha1"},
    extra = s.stringType{
      _H = "ipsecTunEspExtra",
    }
  },

  keylife = s.integerType{
    _H = "ipsecTunKeylife",
    _A = {60, 24*60*60},
    _D = 8*60*60,  -- 8 hours - default Openswan
  },

  ikelifetime = s.integerType{
    _H = "ipsecTunIkelifetime",
    _A = {60, 24*60*60},
    _D = 1*60*60,   -- 1 hour default Openswan
  },

  dpddelay = s.integerType{
    _H = "ipsecTunDpddelay",
    _A = {10, 60*60},
    _D = 30,
  },

  dpdtimeout = s.integerType{
    _H = "ipsecTunDpdtimeout",
    _A = {20, 24*60*60},
    _D = 120,
  },

  pfs = s.newScheme{
    _H = "ipsecTunPfs",
    _T = 'choice',
    _A = {"yes", "no"},
    _V = s.choiceVal,
    _D = "no",
  },

  ["extra-options"] = s.newScheme{
    _H = "ipsecTunExtraoptions",
    _P = -100,
    _NAME = s.stringType{
      _H = "ipsecTunExtraoptionsOption",
    }
  },
}
local function nameOrDefault(val)
   if (val == "%default") then
      return val
   end
   return s.isDefaultNameType(val)
end

ipsecScheme = s.newScheme{
  _H = "ipsec",
  _Z = restartIpsec,
  enable = s.booleanType{_H = "ipsecEnable"},
  secrets = secretsScheme,
  setup = setupScheme,
  connections = s.newScheme{
    _H = "ipsecConnections",
    _P = -100,
    _NAME = ipsecConnectionScheme,
    _NAME_TYPE = nameOrDefault,
  },
  restart = RootOrAdminScheme(
    s.newScheme{
      _H = "ipsecRestart",
      _X = function(path)
        local path = s.clonePath(path, -1)
        local res, err = C.applyConf(path, nil, true)
        return res and "IPsec restarted" or (err or "IPsec not restarted")
      end,
      _R = 'user',
      _P = -90,
    }
  ),
  show = s.newScheme{
    _H = "ipsecShow",
    status = s.newScheme{
      _H = "ipsecShowStatus",
      _X = function()
        return('\n'..s.readpopen("/usr/local/sbin/ipsec auto --status 2>&1"))
      end,
      _R = 'user',
    },
    route = s.newScheme{
      _H = "ipsecShowRoute",
      _X = function()
        return('\n'..s.readpopen("netstat -nr 2>&1"))
      end,
      _R = 'user',
    },

    config = s.newScheme{
      _H = "ipsecShowConfig",
      _X = function()
        return('\n'..s.readpopen("cat /etc/ipsec.conf 2>&1"))
      end,
      _R = 'user',
    },
    ["log"] =  s.newScheme{
      _H = "ipsecShowLog",
      _X = function(path)
        path = s.clonePath(path, -2)
        return C.getDaemonLog(path)
      end,
      _R = 'user',
    },
  }
}
