module(..., package.seeall)
local s = require'scheme'
local iface_mod = require'nsgifaceconf'
local conflib = require'nsgconflib'

local iptables = "/usr/local/sbin/iptables"

local function getProtocols()
  local protocols = conflib.getFile"/etc/protocols"
  protocols = '\n'..(protocols or '')
  local res = {'all'}
  for proto in protocols:gmatch"\n%s*([^%s#]+)%s+%d+%s+" do
    res[#res+1] = proto
    res[#res+1] = '!'..proto
  end
  return res
end

local function getIcmpType()
  local help = s.readpopen(iptables.." -p icmp -h")
  local res = {}
  for it in help:match"Valid ICMP Types:(.*)$":gmatch"[%w%-]+" do
    res[#res+1] = it
    res[#res+1] = '!'..it
  end
  return res
end

local criterionPriority = {
  "protocol",
  "protocol-num",
  "source",
  "destination",
  "in-interface",
  "out-interface",
  "source-port",
  "destination-port",
  "icmp-type",
  "extra-options",
  "target",
  "to-source",
  "to-destination",
  "extra-target",
}
local criterionPatterns = {
  protocol = "-p %s",
  ["protocol-num"] = "-p %d",
  source = "-s %s",
  destination = "-d %s",
  ["in-interface"] = "-i %s",
  ["out-interface"] = "-o %s",
  ["source-port"] = "--sport %s",
  ["destination-port"] = "--dport %s",
  ["icmp-type"] = "--icmp-type %s",
  ["extra-options"] = "%s",
  target = "-j %s",
  ["to-source"] = "--to-source %s",
  ["to-destination"] = "--to-destination %s",
  ["extra-target"] = "%s",
}

local function splitVal(val)
  if (tostring(val)):sub(1,1) == '!' then
    return val:sub(2), '! '
  else
    return val, ""
  end
end

local function setFilterRulesAux(tbl, chain, conf)
  local err = {}
  local function x(com)
    com = " -t "..tbl.." "..com
    local res = s.readpopen(iptables..com.." 2>&1")
    if not res:match'^%s*$' then
      err[#err+1] = '\niptables '..com..'\n'..res:gsub("\nTry[^\n]*\n$", '')
    end
  end
  x("-F "..chain)
  if #err > 0 then
    return false, err[1]
  end
  --for _, rule in ipairs(conf) do 
  for idx = 0, #conf do
    local rule = conf[idx]
    if rule then
      local com = {"-A "..chain}
      for _, crit in ipairs(criterionPriority) do
        local val = rule[crit]
        if crit == "in-interface" or crit == "out-interface" then
          -- may be interface aliases
          val = iface_mod.getIfaceName(val)
        end
        local pat = criterionPatterns[crit]
        if val and pat then
          local val, prefix = splitVal(val)
          com[#com+1] = prefix..pat:format(val)
        end
      end
      x(table.concat(com, ' '))
    end
  end
  if conf["default-target"] then
    x("-P "..chain..' '..conf["default-target"])
  end
  if #err > 0 then
    return false, table.concat(err)
  end
  return true
end

local filterAliasHookKey = {
  nat = {
    PREROUTING = {},
    POSTROUTING = {},
    OUTPUT = {},
  },
  filter = {
    FORWARD = {},
    INPUT = {},
    OUTPUT = {},
  },
}
local function setFilterRules(path, sch, conf, prevConf)
  if not s.confCmp(conf, prevConf) then
    s.confValidate(path, sch, conf, true) -- may be bad parms from previos conf
    conf = s.addDefault(path, sch, conf, true)
    local tbl = path[#path-1]
    local chain = path[#path]
    iface_mod.setAliasHook(filterAliasHookKey[tbl][chain], 
      conf[1] and
      function()
        return  setFilterRulesAux(tbl, chain, conf)
      end or nil
    )
    return setFilterRulesAux(tbl,chain, conf)
  end
  return true
end

local portRange = s.newScheme{
  _T = 'string',
  _V = function(val)
    if type(val) == 'number' and val >= 0 and val <= 0xffff and val%1==0 then
      return tostring(val)
    end
    if type(val) == 'string' then
      local negative, p1, p2 = val:match"^(!?)(%d%d?%d?d?%d?)(:?%d?%d?%d?d?%d?)$"
      if (
        p1 and tonumber(p1) <= 0xffff  and
        (p2=='' or p2:sub(1,1) == ':' and tonumber(p2:sub(2)) <= 0xffff)
      ) then
        return val
      end
    end
    return nil, "port range in format  port[:port]"
  end
}

local criterionScheme =s.newScheme{
  _H = "nfFilterRule",
  protocol = s.newScheme{
    _H = "nfCriterionProtocol",
    _T = "choice",
    _A = getProtocols(),
    _V = s.choiceVal,
    _P = 60,
    _DEP = 1,
  },
  ["protocol-num"] = function(_, path)
    return not C.getConfOrDefault(s.clonePath(path, -1, "protocol")) and
    s.integerType{
      _H = "nfCriterionProtocolNum",
      _A = {0, 0xffff},
      _P = 50,
    }
  end,
  ["in-interface"] = s.stringType{
    _H = "nfCriterionInInterface",
    _P = 40,
  },
  ["out-interface"] = s.stringType{
    _H = "nfCriterionOutInterface",
    _P = 30,
  },
  source = s.stringType{
    _H = "nfCriterionSource",
    _P = 20
  },
  destination = s.stringType{
    _H = "nfCriterionDestination",
    _P = 10,
  },
  ["source-port"] = function(_, path)
    local proto = C.getConfOrDefault(s.clonePath(path, -1, "protocol"))
    return (proto == 'tcp' or proto == 'udp') and
    portRange{
      _H = "nfCriterionSport",
      _P = 8,
    }
  end,
  ["destination-port"] = function(_, path)
    local proto = C.getConfOrDefault(s.clonePath(path, -1, "protocol"))
    return (proto == 'tcp' or proto == 'udp') and
    portRange{
      _H = "nfCriterionDport",
      _P = 7,
    }
  end,
  ["icmp-type"] = function(_, path)
    local proto = C.getConfOrDefault(s.clonePath(path, -1, "protocol"))
    return (proto == 'icmp') and
    s.newScheme{
      _H = "nfCriterionIcmpType",
      _T = "choice",
      _A = getIcmpType(),
      _V = s.choiceVal,
      _P = 7,
    }
  end,
  ["extra-options"] = s.stringType{
    _H = "nfCriterionExtra",
    _P = 1,
  },
  ["extra-target"] = s.stringType{
    _H = "nfExtraTarget",
    _P = -100,
  },
}

local filterCriterionParms  = criterionScheme{
  target = s.newScheme{
    _H = "nfFilterTarget",
    _T = "choice",
    _A = {
      "ACCEPT",
      "REJECT",
      "DROP",
      "RETURN"
    },
    _V = s.choiceVal,
  },
}

local filterChainScheme = s.newScheme{
  _Z = setFilterRules,
  _LIST = filterCriterionParms,
  ["default-target"] = s.newScheme{
    _H = "nfFilterDefaultTarget",
    _T = "choice",
    _A = {
      "ACCEPT",
      "DROP",
    },
    _V = s.choiceVal,
  },
}

local LO_ADDRESS = "127.0.0.1"
local loAddrScheme = s.stringType{
  _H = "nfCriterionSource",
  _V = function(val)
    return val == LO_ADDRESS and LO_ADDRESS or
    nil, LO_ADDRESS
  end,
  _D = LO_ADDRESS,
}
local onlyAcceptTargetScheme = s.newScheme{
  _H = "nfFilterTarget",
  _T = "choice",
  _A = {
    "ACCEPT",
  },
  _V = s.choiceVal,
  _D = "ACCEPT",
}
local filterInput0 = s.newScheme{
  _H = "nfFilterInputRule0",
  source = loAddrScheme,
  target = onlyAcceptTargetScheme,
}
local filterOutput0 = s.newScheme{
  _H = "nfFilterOutputRule0",
  destination = loAddrScheme,
  target = onlyAcceptTargetScheme,
}

filterScheme = s.newScheme{
  _H = "nfFilter",
  _Z = function() return true end,
  INPUT =filterChainScheme{
    _H = "nfFilterInput",
    [0] = filterInput0,
    _LIST = filterCriterionParms{
      ["out-interface"] = s.nilType,
    },
    _P = 30,
  },
  FORWARD = filterChainScheme{
    _H = "nfFilterForward",
    _P = 20,
  },
  OUTPUT =  filterChainScheme{
    _H = "nfFilterOutput",
    [0] = filterOutput0,
    _LIST = filterCriterionParms{
      ["in-interface"] = s.nilType,
    },
    _P = 10,
  },
  show = s.newScheme{
    _H = "nfFilterShow",
    _X = function()
      return('\n'..s.readpopen(iptables.."  -nvL -t filter 2>&1"))
    end,
    _R = 'user',
  }
}

--------------------------------------------------------------------------------
-- NAT

local natCriterionParms  = criterionScheme{
  _H = "nfNatRule",
  target = s.newScheme{
    _H = "nfNatTarget",
    _T = "choice",
    _A = {
      "DNAT",
      "REDIRECT",
      "NETMAP",
    },
    _V = s.choiceVal,
    _DEP = 1,
  },
  ["to-destination"] = function(_, path)
    local targ = C.getConfOrDefault(s.clonePath(path, -1, "target"))
    return (targ == 'DNAT') and
    s.stringType{
      _H = "nfNatToDestination",
      _P = -10,
    }
  end,

}

local natChainScheme = s.newScheme{
  _Z = setFilterRules,
  _LIST = natCriterionParms,
  ["default-target"] = s.newScheme{
    _H = "nfNatDefaultTarget",
    _T = "choice",
    _A = {
      "ACCEPT",
      "DROP",
    },
    _V = s.choiceVal,
  },
}

natScheme = s.newScheme{
  _H = "nfNat",
  _Z = function() return true end,

  PREROUTING =  natChainScheme{
    _H = "nfNatPrerouting",
    _LIST = natCriterionParms{
      ["out-interface"] = s.nilType,
    },
    _P = 30,
  },
  POSTROUTING =  natChainScheme{
    _H = "nfNatPostrouting",
    _LIST = natCriterionParms{
      ["in-interface"] = s.nilType,
      target = s.newScheme{
        _H = "nfNatPostroutingTarget",
        _T = "choice",
        _A = {
          "SNAT",
          "MASQUERADE",
        },
        _V = s.choiceVal,
        _DEP = 1,
      },
      ["to-source"] = function(_, path)
        local targ = C.getConfOrDefault(s.clonePath(path, -1, "target"))
        return (targ == 'SNAT') and
        s.stringType{
          _H = "nfNatToSource",
          _P = -10,
        }
      end,
    },
    _P = 20,
  },
  OUTPUT =  natChainScheme{
    _H = "nfNatOutput",
    _LIST = natCriterionParms{
      ["in-interface"] = s.nilType,
    },
    _P = 10,
  },
  show = s.newScheme{
    _H = "nfNatShow",
    _X = function()
      return('\n'..s.readpopen(iptables.."  -nvL -t nat 2>&1"))
    end,
    _R = 'user',
  }
}

--------------------------------------------------------------------------------

showNetFilter = s.newScheme{
  _H = "nfShow",
  _X = function()
    return('\n'..s.readpopen(iptables.."-save 2>&1"))
  end
}

nfScheme = s.newScheme{
  _H = "nf",
  _Z = function() return true end,
  filter = filterScheme,
  nat = natScheme,
  show = showNetFilter,
}
