module(..., package.seeall)
local s = require'scheme'
local iface_mod = require'nsgifaceconf'
local conflib = require'nsgconflib'
local sms = require'nsgsmsconf'
local usb = require'nsgusbconf'

local authName = {
  chap = "chap",
  msChap = "ms-chap",
  msChapV2 = "ms-chap-v2",
  pap = "pap",
  eap = "eap",
}

local ipUpScript = [[
#!/usr/bin/lua
local dns1 = os.getenv'DNS1'
if dns1 then
  dns1 = dns1:gsub('%.', '..')
  local dns2 = os.getenv'DNS2'
  if dns2 then
    dns2 = ',\\"'..dns2:gsub('%.', '..')..'\\"'
  else
    dns2 = ""
  end
  local ifname = os.getenv'IFNAME'
  os.execute(
    '/usr/sbin/nsgsh -q system.dns-client.add-dyn-nameservers.{\\"'..ifname..
    '\\",\\"'..dns1..'\\"'..dns2..'}'
  )
end
]]

local ipDownScript = [[
#!/usr/bin/lua
local ifname = os.getenv'IFNAME'
os.execute(
  '/usr/sbin/nsgsh -q system.dns-client.add-dyn-nameservers.{\\"'..ifname..'\\"}'
)
]]

local ox = require'posix'
local function addIpUpDownScripts()
  local ipup = '/etc/ppp/ip-up'
  --[[
  local f = io.open(ipup)
  if f then 
    f:close()
    return
  end
--]]
  f = io.open(ipup, 'w')
  f:write(ipUpScript)
  f:close()
  ox.chmod(ipup, 'a+x')
  local ipdown = '/etc/ppp/ip-down'
  f = io.open(ipdown, 'w')
  f:write(ipDownScript)
  f:close()
  ox.chmod(ipdown, 'a+x')
end

local cdmaScriptPattern = "TIMEOUT %d XXX-\\rAT-OK ATD#777 "
.."CONNECT ''"
local gprsScriptPattern = "TIMEOUT %d XXX-\\rAT-OK AT+CGDCONT=1,\"IP\",\"%s\" "
.."OK ATD*99***1# CONNECT ''"

local function getChatScript(conf, typeOfPort)
  if conf then
    return conf.script 
    or (typeOfPort == 'cdma' and cdmaScriptPattern or gprsScriptPattern):format(
      conf.timeout or 30, conf.APN or ""
    )
  else
    return ''
  end
end
local function getChatCall(conf, typeOfPort)
  local f = io.open(conf._tmpFileName, 'w')
  f:write(getChatScript(conf, typeOfPort))
  f:close()
  return ("chat -v -s -f %s"):format(conf._tmpFileName)
end

local auth2opt = {
  [authName.pap] = "require-pap",
  [authName.chap] = "require-chap",
  [authName.msChap] = "require-mschap",
  [authName.msChapV2] = "require-mschap-v2",
}
local mppe2opt = {
  auto = "require-mppe",
  [40] = "require-mppe-40",
  [128] = "require-mppe-128",
}

local function addCondOptionFunc(key, opt)
  return function(res, val)
    res[#res+1] = val == key and opt or nil
  end
end
local function addOptionFunc(opt)
  return addCondOptionFunc(true, opt)
end
local function addOptionValFunc(opt)
  return function(res, val)
    res[#res+1] = opt
    res[#res+1] = val
  end
end

local addPPPDOpt addPPPDOpt = {
  authentication = function(res, val) 
    res[#res+1] = auth2opt[val] 
  end,
  connection = function(res, val) 
    if val == 'on-demand' then
      res[#res+1] = "demand"
      res[#res+1] = "nopersist"
    elseif val == 'passive' then
      res[#res+1] = "silent"
    end
  end,
  ["data-compression"] = function(res, val) 
    res[#res+1] = val ~= true and 'noccp' or nil
  end,
  ["encrypt-mppe"] = function(res, val) 
    res[#res+1] = mppe2opt[val] or 'nomppe'
  end,
  ["encrypt-mppe-mode"] = function(res, val) 
    if val == 'stateful' then
      res[#res+1] = "mppe-stateful"
      res[#res+1] = "require-mppe-stateful" 
    end
  end,

  ["debug-level"] = addCondOptionFunc(1, 'debug'),
  ["sent-username"] = addOptionValFunc('name'),
  ["session-time"] = addOptionValFunc('maxconnect'),
  ["default-route"] = addOptionFunc('defaultroute'),
  --["default-route-metric"] = addOptionValFunc('metric'),
  ["lcp-echo-interval"] = addOptionValFunc('lcp-echo-interval'),
  ["lcp-echo-failure"] = addOptionValFunc('lcp-echo-failure'),
  ["mtu"] = addOptionValFunc('mtu'),
  ["idle-time"] = addOptionValFunc('idle'),

  --------------------------------------------------------------------------------
  ["ipcp"] = function(res, val) 
    for k, v in pairs(val) do
      addPPPDOpt['ipcp.'..k](res, v)
    end
  end,
  ["ipcp.accept-address"] = addOptionFunc('ipcp-accept-local'),
  ["ipcp.accept-peer-address"] = addOptionFunc('ipcp-accept-remote'),
  ["ipcp.accept-dns"] = addOptionFunc('usepeerdns'),
  ["ipcp.dns"] = addOptionValFunc('ms-dns'),

  --------------------------------------------------------------------------------
  ["refuse-auth"] = function(res, val) 
    for k, v in pairs(val) do
      addPPPDOpt['refuse-auth.'..k](res, v)
    end
  end,
  ["refuse-auth."..authName.chap] = addOptionFunc('refuse-chap'),
  ["refuse-auth."..authName.eap] = addOptionFunc('refuse-eap'),
  ["refuse-auth."..authName.msChap] = addOptionFunc('refuse-mschap'),
  ["refuse-auth."..authName.msChapV2] = addOptionFunc('refuse-mschap-v2'),
  ["refuse-auth."..authName.pap] = addOptionFunc('refuse-pap'),
  --------------------------------------------------------------------------------

  ["chat"] = function(res, val, path)
    res[#res+1] = "connect"
    res[#res+1] = getChatCall(val, usb.getOptionRecursive(path, 'wirelessType'))
  end
}
setmetatable(addPPPDOpt, {__index = function()return function()end end})


local pppUnitNumber = 0
local function getUnit(path)
  local unit = C.getOption(path, 'pppUnit')
  if not unit then
    pppUnitNumber = pppUnitNumber + 1
    unit = pppUnitNumber
    C.setOption(path, 'pppUnit', unit)
    iface_mod.addIfaceAlias(path[#path], 'ppp'..tostring(unit))
  end
  return unit
end

function getPPPParams(path, portConf, channel, nopreamble)
  channel = channel or 'ppp'
  local ttyDev = usb.port2tty[path[#path]]
  local res  = nopreamble or {
    'pppd', 
    'nodetach', 
  }
  local ttypos, setrealtty 

  if not nopreamble then
    local addr_block
    if portConf.ppp[channel]["ip-address"] then
      addr_block = portConf.ppp[channel]["ip-address"] .. ":"
    end
    if portConf.ppp[channel]["peer-ip-address"] then
      addr_block = (addr_block or ":") ..  portConf.ppp[channel]["peer-ip-address"]
    end
    res[#res+1] = addr_block
  end

  res[#res+1] = 'unit'
  res[#res+1] = getUnit(path)

  local smsType = C.getOption(path, "smsType")

  if (smsType == "+++" and 
    portConf['sms-handler'] and 
    portConf['sms-handler']['ppp-cooperation']) then

    res[#res+1] = 'pty'
    local nsgsmsCommand, ttyParPosition = 
    sms.getNsgsmsCom(ttyDev, portConf['sms-handler'])

    nsgsmsCommand[#nsgsmsCommand+1] = '-p'
    nsgsmsCommand[#nsgsmsCommand+1] = 'ppp'
    res[#res+1] = table.concat(nsgsmsCommand, ' ')
    ttypos = #res        --++ position number
    setrealtty = function(oper, realtty)
      nsgsmsCommand[ttyParPosition] = realtty
      oper.com[ttypos] = table.concat(nsgsmsCommand, ' ')
    end
  else
    res[#res+1] = ttyDev -- real ttyName or ''
    ttypos = #res        --++ position number
    setrealtty = function(oper, realtty)
      oper.com[ttypos] = realtty
    end
    if portConf.baudrate then
      res[#res+1] = tostring(portConf.baudrate)
    end
  end

  for k, v in pairs(portConf.ppp[channel]) do
    addPPPDOpt[k](res, v, path)
  end

  -- extra options
  local extOpt = portConf.ppp[channel].options
  if type(extOpt) == 'string' then
    for _, v in ipairs(conflib.splitArgs(extOpt)) do
      res[#res+1] = v
    end
  end
  return res, setrealtty   --++ add position
end



local function resetModem(oper)
   local path = oper.options.extra.path
  C.log_warn("Modem seems to be frozen, RESET.")
  usb.powerOff(path)
end

local pppdExitCode = {
  [-1] = true,
  [5]  = true,
  [12] = true,
  [13] = true,
}

local function pppAtInit(oper)
  local path = oper.options.extra.path    --++
  usb.setGPIO(path)
  local sim = oper.options.extra.mainCom and usb.mainSimNumber or usb.auxSimNumber
  usb.setSim(path, sim)
  oper.stat.sim = sim == usb.mainSimNumber and 'MAIN' or 'AUX'
end

local function pppAtStart(oper)
  oper.atRead = oper.options.atRead -- return to first atRead function
  oper.extra.myLog = ""

  local path = oper.options.extra.path    --++
  usb.togglePower(path)
  C.threadSleep(5)
  for i = 1, 30 do
    if usb.probeUSBDeviceScheme(nil, path) then  --++        
      oper.options.extra.realtty(oper, usb.port2tty[path[#path]])
      break
    end
    C.threadSleep(1)
  end
end

local function pppAtStartOnlyTimeout(oper)
  C.threadSleep(5)
end

local function pppAtRead3(oper)
  oper.extra.myLog = (oper.extra.myLog..oper.log:read(-1)):sub(-1024)
  if oper.extra.myLog:match"Modem hangup" then
    oper.atRead = nil
    oper.extra.myLog = nil
    C.timer:del(oper.extra.hangupTimer)
  end
end

local function pppAtRead2(oper)
  oper.extra.myLog = (oper.extra.myLog..oper.log:read(-1)):sub(-1024)
  if oper.extra.myLog:match"Connection terminated" then
    oper.atRead = pppAtRead3
    oper.extra.hangupTimer = C.timer:add(10, resetModem, oper)
  end
end

local function pppAtRead(oper)
  oper.extra.myLog = (oper.extra.myLog..oper.log:read(-1)):sub(-1024)
  local iface = oper.extra.myLog:match"Using interface (ppp%d+)"
  if iface then
    oper.stat.iface = iface 
    oper.atRead = pppAtRead2
    --oper.extra.myLog = nil
    --[[
    if oper.options.extra.mainNat and oper.com == oper.options.extra.mainCom
    then
      setNat(iface)
      oper.extra.nat = true
    elseif oper.options.extra.auxNat and oper.com == oper.options.extra.auxCom 
    then
      oper.extra.nat = true
      setNat(iface)
    end
    ]]
  end
end

local function pppAtExit(oper)
  C.timer:del(oper.extra.hangupTimer)
  if oper.stat.iface then
    --[[
    if oper.extra.nat then
      unsetNat(oper.stat.iface)
    end
    ]]
    oper.stat.iface = nil 
  end
  if not oper.options.extra.auxCom or not oper.options.extra.mainCom then
    return
  end
  oper.extra.count = (oper.extra.count or 0) + 1
  local badExit = not pppdExitCode[oper.exitCode.exitstatus or -1]
  local path = oper.options.extra.path --++
  if oper.com == oper.options.extra.mainCom then
    if badExit and oper.extra.count < oper.options.extra.mainCount then
      return
    elseif badExit then
      oper.com = oper.options.extra.auxCom
      usb.setSim(path, auxSimNumber)
      oper.stat.sim = 'AUX'
    end
  else
    if badExit and oper.extra.count < oper.options.extra.auxCount then
      return
    elseif badExit or oper.options.extra.priority then
      oper.com = oper.options.extra.mainCom
      usb.setSim(path, usb.mainSimNumber)
      oper.extra.count = 0
      oper.stat.sim = 'MAIN'
    end
  end
  oper.extra.count = 0
end

local function fcon(...)
  local t = {...}
  local n = select('#', ...)
  return function(...)
    for i = 1, n do
      if t[i] then
        t[i](...)
      end
    end
  end
end


function restartPPPDaemon(path, sch, conf)
  addIpUpDownScripts()
  --s.confValidate(path, sch, conf, true)
  local pppconf = s.addDefault(path, sch, conf, true) --clone conf with defaults
  local tmpFileList = {}
  local realtty
  local mainCom
  if pppconf.ppp.main.attempts > 0 then
    if pppconf.ppp.main.chat then
      pppconf.ppp.main.chat._tmpFileName = os.tmpname()
      tmpFileList[#tmpFileList+1] = pppconf.ppp.main.chat._tmpFileName
    end
    mainCom, realtty = getPPPParams(path, pppconf, 'main')
  end
  local auxCom
  if pppconf.ppp.aux and pppconf.ppp.aux.attempts > 0 then
    if pppconf.ppp.aux.chat then
      pppconf.ppp.aux.chat._tmpFileName = os.tmpname()
      tmpFileList[#tmpFileList+1] = pppconf.ppp.aux.chat._tmpFileName
    end
    auxCom, realtty = getPPPParams(path, pppconf, 'aux')
  end
  local wls = usb.isWirelessPort[path[#path]]
  local dualSim = usb.getOptionRecursive(path, 'simNumber')
  local powered = usb.getOptionRecursive(path, 'controlledPower')

  -- prepare sms-handler
  local startSmsHandlerOnAuxTty, stopSmsHandlerOnAuxTty
  local smsType = usb.getOptionRecursive(path, 'smsType')
  if smsType == 'aux' and pppconf['sms-handler']['ppp-cooperation'] then
    local ttyDev = usb.port2tty[path[#path]]
    local smsCom, smsttypar = sms.getNsgsmsCom(ttyDev, pppconf['sms-handler'])
    function startSmsHandlerOnAuxTty(oper)  
      local path = oper.options.extra.path
      C.restartDaemon(
        oper.options.extra.smsPath, 
        smsCom,
        {
          atStart = function(oper)
            local tty0 = usb.port2tty[path[#path]]
            local tty2 = tty0:gsub("(%d+)$", 
              function(s) return tostring(tonumber(s) + 2) end)
            oper.com[smsttypar] = tty2
          end
        }
      )
    end  
    function stopSmsHandlerOnAuxTty(oper)  
      C.restartDaemon(oper.options.extra.smsPath)
    end  
  end
  --

  C.restartDaemon(
    path, 
    mainCom or auxCom,
    {
      tmpFileList = tmpFileList,

      atInit = dualSim and pppAtInit,
      atStart = fcon(powered and pppAtStart or pppAtStartOnlyTimeout, startSmsHandlerOnAuxTty),
      atRead = wls and pppAtRead,
      atExit = fcon(stopSmsHandlerOnAuxTty, dualSim and pppAtExit),
      extra = {
        path = path,  --++
        smsPath = s.clonePath(path, -1, path[#path].."_SMS"),
        realtty = realtty,  --++
        mainCom = mainCom,
        auxCom = auxCom,
        mainCount = pppconf.ppp.main.attempts or 1,
        auxCount = pppconf.ppp.aux and pppconf.ppp.aux.attempts or 1,
        priority = pppconf.ppp.priority,
        --[[
        mainNat = pppconf.ppp.main.nat,
        auxNat = pppconf.ppp.aux and pppconf.ppp.aux.nat,
        ]]
      },
    }
  )
end


local function isCdmaModule(path)
  local modName =  usb.getOptionRecursive(path, 'wirelessType')
  return modName == "cdma"
end
local function isGprsModule(path)
  local modName = usb.getOptionRecursive(path, 'wirelessType') 
  return modName == "3g" or modName == "edge"
end
local chatScriptScheme = s.newScheme{
  _H = "pppChatScript",
  _N = function(sch, conf)
    if conf and conf.script then
      return {true, script = true}
    end
    return {}
  end,
  script = s.newScheme{
    _H = "pppChatScriptScript",
    _T = "string",
    _V = s.stringVal,
    _P = 100,
  },
  show = s.newScheme{
    _H = "pppChatScriptShow",
    _P = -100,
    _X = function(path, sch, conf)
      path = s.clonePath(path, -1) 
      porttype = usb.getOptionRecursive(path, 'wirelessType')                        
      return getChatScript(C.getConfByPath(path), porttype)  
    end,
    _R = 'user',
  },
  timeout = function(_, path)
    return ((isGprsModule(path) or isCdmaModule(path)) and
      s.newScheme{
        _H = "pppChatScriptTimeout",
        _T = "number",
        _A = {1, 86400},
        _V = s.integerVal,
        _D = 30,
      }
    )
  end,
  APN = function(_, path)
    return ((isGprsModule(path)) and
      s.newScheme{
        _H = "pppChatScriptApn",
        _T = "string",
        _V = s.stringVal,
        _D = "",
      }
    )
  end,
}

local attemptsScheme = s.newScheme{
  _H = "pppAttempts",
  _T = "number",
  _A = {0, 2^31},
  _V = s.integerVal,
  _D = 1,
}

local pppParmsScheme = s.newScheme{
  _H = "ppp",
  attempts = attemptsScheme,
  authentication = s.newScheme{
    _H = "pppAuthentication",
    _T = 'choice',
    _A = {
      {authName.chap, "Challenge Handshake Authentication Protocol(CHAP)"},
      {authName.msChap, "Microsoft CHAP(MS-CHAP"},
      {authName.msChapV2, "Microsoft CHAP Version 2(MS-CHAP-V2)"},
      {authName.pap, "Password Authentication Protocol(PAP)"},
      {"noauth", "Without any authentication procedure"},
    },
    _V = s.choiceVal,
    _D = "noauth",
  },

  connection = s.newScheme{
    _H = "pppConnection",
    _T = 'choice',
    _A = {
      {"on-demand", "initiate the link only when data traffic is present"},
      --{"passive", "stay inactive until a LCP-request has been received"},
      {"permanent", "maintain a connection all the time"},
    },
    _V = s.choiceVal,
    _D = "permanent",
  },
  ["data-compression"] = s.booleanType{
    _H = "pppDataCompression",
    _D = false,
  },
  ["debug-level"] = s.newScheme{
    _H = "debugLevel",
    _T = "number",
    _A = {0, 1},
    _V = s.integerVal,
    _D = 0,
  },

  ["encrypt-mppe"] = s.newScheme{
    _H = "pppEncryptMppe",
    _T = 'choice',
    _A = {
      {128, "128 Bit Encryption only"},
      {40, "40 Bit Encryption only"},
      {"auto", "Will offer 40 and 128 bit"},
      {"no", "Encryption disable"},
    },
    _V = s.choiceVal,
    _D = "no",
  },
  ["encrypt-mppe-mode"] = s.newScheme{
    _H = "pppEncryptMppeMode",
    _T = 'choice',
    _A = {
      {"stateful", "Stateful mode only"},
      {"stateless"},
    },
    _V = s.choiceVal,
    _D = "stateless",
  },

  ["idle-time"] = s.newScheme{
    _H = "pppIdleTime",
    _T = "number",
    _A = {30, 86400},
    _V = s.integerVal,
  },
  ipcp = s.newScheme{
    _H = "pppIpcp",
    ["accept-address"] = s.booleanType{
      _H = "pppIpcpAcceptAddress",
      _D = true,
    },
    ["accept-peer-address"] = s.booleanType{
      _H = "pppIpcpRemoteAddress",
      _D = true,
    },
    ["accept-dns"] = s.booleanType{
      _H = "pppIpcpAcceptDns",
      _D = true,
    },
    --dns = s.ipAddressType{_H = "pppIpcpDns"},
  },
  options = s.stringType{
    _H = "pppOptions",
  },
  ["refuse-auth"] = s.newScheme{
    _H = "pppRefuseAuth",
    [authName.chap] = s.booleanType{_H = "pppRefuseAuthChap"},
    [authName.eap] = s.booleanType{
      _H = "pppRefuseAuthEap",
      _D = true
    },
    [authName.msChap] = s.booleanType{_H = "pppRefuseAuthMsChap"},
    [authName.msChapV2] = s.booleanType{_H = "pppRefuseAuthMsChapV2"},
    [authName.pap] = s.booleanType{_H = "pppRefusePap"},
  },
  ["sent-username"] = s.stringType{
    _H = "pppSentUsername",
  },
  ["session-time"] = s.newScheme{
    _H = "pppSessionTime",
    _T = "number",
    _A = {60, 86400},
    _V = s.integerVal,
  },
  ["default-route"] = s.booleanType{_H = "pppDefaultRoute"},
  ["default-route-metric"] = s.newScheme{
    _H = "pppDefaultRouteMetric",
    _T = "number",
    _A = {1, 255},
    _V = s.integerVal,
    _D = 1,
  },
  ["chat"] = chatScriptScheme,

  ["lcp-echo-interval"] = s.newScheme{
    _H = "pppLcpEchoInterval",
    _T = "number",
    _A = {0, 86400},
    _V = s.integerVal,
    _D = 10,
  },
  ["lcp-echo-failure"] = s.newScheme{
    _H = "pppLcpEchoFailure",
    _T = "number",
    _A = {0, 255},
    _V = s.integerVal,
    _D = 3,
  },
  ["ip-address"] = s.ipAddressType{_H = "pppIpAddress"},
  ["peer-ip-address"] = s.ipAddressType{
    _H = "pppPeerIpAddress",
  },
  mtu = s.newScheme{
    _H = "pppMtu",
    _T = "number",
    _A = {100, 1500},
    _V = s.integerVal,
    _D = 1500,
  },
  --[[
  nat =  s.booleanType{
    _H = "pppNat",
    _D = false,
  },
  ]]
}

local function getPppNatRule(path)
  local rule = {
    ["out-interface"] = path[#path-2],
    target = "MASQUERADE",
    _keep = true,
  }
  local natPath = {'ip', 'nat', 'POSTROUTING'}
  local natPostroutingList = C.getConfByPath(natPath)
  if natPostroutingList then
    for k, v in ipairs(natPostroutingList) do
      if s.confCmp(rule, v) then
        return true, natPath, k, rule
      end
    end
    return false, natPath, #natPostroutingList + 1, rule
  else
    return false, natPath, 1, rule
  end
end

function isDualSimModule(modName)
  return 
  modName == "3g" or
  modName == "edge"
end
pppScheme = s.newScheme{
  _H = "ppp",
  main = pppParmsScheme{_H = "pppMain", _P = 100},
  aux = function(_, path)
    return isDualSimModule(path[#path - 2]) and
    pppParmsScheme{
      _H = "pppAux", 
      _P = 99,
      attempts = attemptsScheme{_D = 0},
    }
  end,
  priority = function(_, path)
    return isDualSimModule(path[#path - 2]) and
    s.booleanType{_H = "pppPriority", _P = 98}
  end,

  ["add-nat"] = function(_, path)
    return not getPppNatRule(path) and
    s.newScheme{
      _H = "pppAddNat",
      _X = function(path)
        local res, natPath, num, rule = getPppNatRule(path)
        if res then   
          return "Rule already exists: "..table.concat(natPath, '.')..'.'..num
        else
          C.setConfigValue(s.clonePath(natPath, nil, num), rule)
          C.applyConf(natPath)
          return "New rule add to "..table.concat(natPath, '.')..'.'..num
        end
      end,
      _DEP = 4,
    }
  end,

  ["del-nat"] = function(_, path)
    return getPppNatRule(path) and
    s.newScheme{
      _H = "pppAddNat",
      _X = function(path)
        local res, natPath, num, rule = getPppNatRule(path)
        if res then
          C.removeConfigValue(s.clonePath(natPath, nil, num))
          C.applyConf(natPath)
          return "Rule deleted: "..table.concat(natPath, '.')..'.'..num
        else
          return "No rule to delete in "..table.concat(natPath, '.')
        end
      end,
      _DEP = 4,
    }
  end,

  ["log"] =  s.newScheme{
    _H = "pppLog",
    _X = function(path)
      path = s.clonePath(path, -2)
      return C.getDaemonLog(path)
    end,
    _R = 'user',
  },

  show = s.newScheme{
    _H = "pppShow",
    _X = function(path)
      path = s.clonePath(path, -2)
      local stat = C.getDaemonStat(path) or {}
      local tty = usb.port2tty[path[#path]]
      local out = "\n"
      .."tty device: "..tty.."\n"
      local sim = stat.sim or 'unknown'
      out = out.."SIM:        "..sim.."\ninterface:  "
      local unit = C.getOption(path, 'pppUnit')
      if unit then
        out = out..'\n'..iface_mod.showInterfaceStat('ppp'..unit)
      else
        out = out.."No PPP network interface\n"
      end
      --[[
       local iface = stat.iface
      if iface then
        out = out..'\n'..iface_mod.showInterfaceStat(iface)
      else
        out = out.."No PPP network interface\n"
      end
]]
      return out
    end,
    _R = 'user',
  },
}
--------------------------------------------------------------------------------
-- secrets stuff

local function createPPPsecretsFile(filename, conf)
  conf = conf or {}
  local secrets = ""
  for _, v in ipairs(conf) do
    local ipaddr = ""
    for _, ip in ipairs(v["ip-addresses"]or{}) do
      ipaddr = ipaddr..' '..ip
    end
    secrets = secrets..(("%s %s %s %s"):format(
        v.client or '*', v.server or '*', v.secret or '""', ipaddr 
    ))..'\n'
  end
  local f = io.open(filename, 'w')
  if f then
    f:write(secrets)
    f:close()
  end
end

local function createPppSecrets(path, sch, conf)
  conf = s.addDefault(path, sch, conf, true) --clone conf with defaults
  createPPPsecretsFile("/etc/ppp/pap-secrets", conf.pap)
  createPPPsecretsFile("/etc/ppp/chap-secrets", conf.chap)
  return true
end

local pppsecretsScheme = s.newScheme{
  _LIST = s.newScheme{
    client = s.stringType{
      _H = "pppSecretsClient",
      _D = '*',
    },
    server = s.stringType{
      _H = "pppSecretsServer",
      _D = '*',
    },
    secret = s.stringType{
      _H = "pppSecretsSecret",
      _D = '""',
    },
    ["ip-addresses"] = s.newScheme{
      _H = "pppSecretIpaddresslist",
      _LIST = s.stringType{
        _H = "pppSecretIpaddress",
      },
    }
  }
}

secretsScheme = s.newScheme{
  _H = "pppsecrets",
  _Z = createPppSecrets,
  pap = pppsecretsScheme{
    _H = "pppSecretsPap",
  },
  chap = pppsecretsScheme{
    _H = "pppSecretsChap",
  },
}

startRawAccess = function(path, sch, conf, sesNum)
  local ttyDev = usb.port2tty[path[#path-2]]
  local sim = C.getConfOrDefault(s.clonePath(path, -1, "sim"))
  sim = sim == 'aux' and usb.auxSimNumber or usb.mainSimNumber
  C.restartDaemon(s.clonePath(path, -2)) -- kill ppp daemon if it is
  C.restartDaemon(path, {'nsgcu', ttyDev}, 
    {
      sesNum = sesNum, noTimestamp = true, noRestart = true,
      atInit = function(oper)
        local path = oper.options.extra.path
        oper.log:write("Port initialization, wait... ")
        usb.setGPIO(path)
        usb.setSim(path, sim)
      end,
      atStart = function(oper) 
        local path = oper.options.extra.path
        usb.togglePower(path)
        C.threadSleep(5)
        for i = 1, 30 do
          if usb.probeUSBDeviceScheme(nil, path) then  --++ 
            oper.options.extra.realtty(oper, usb.port2tty[path[#path]])
            break
          end
          C.threadSleep(1)
        end
        oper.log:write("OK\r\n")
      end,
      extra = {
        path = s.clonePath(path, -2),  --++ go to port
        realtty = function(oper, realtty)
          oper.com[2] = realtty
        end
      },
    }
  ) 
end

rawAccessScheme = s.newScheme{
  _H = "portRawAccessNode",
  launch = s.newScheme{
    _H = "portRawAccess",
    _X = startRawAccess,
    _XI = true,
  },
  sim = function(_, path)
    return isDualSimModule(path[#path - 2]) and
    s.newScheme{
      _H = "portRawAccessSim",
      _T = 'choice',
      _A = {"main", "aux"},
      _V = s.choiceVal,
      _D = "main",
    }
  end,
}
