module(..., package.seeall)
local s = require'scheme'
local ppp = require'nsgpppconf'
local soft = require'nsgsoftware'
local nsgconfig = require'nsgconfig'
local dns = require'nsgdnsconf'
local iface_mod = require'nsgifaceconf'

local function getOutOfUtil(util)
  return function()
    return('\n'..s.readpopen(util.." 2>&1"))
  end
end

local function getDateTable(str)
  local y, M, d, h, m, s = str:match'^%s*(%d%d%d%d)-(%d%d)-(%d%d)%s+(%d%d):(%d%d):(%d%d)%s*$'
  if not y then
    y, M, d, h, m = str:match'^%s*(%d%d%d%d)-(%d%d)-(%d%d)%s+(%d%d):(%d%d)%s*$'
    if not y then
      h, m, s = str:match'^%s*(%d%d):(%d%d):(%d%d)%s*$'
      if not h then
        h, m = str:match'^%s*(%d%d):(%d%d)%s*$'
      end
    end
  end
  if h then
    return {
      year = y and tonumber(y),
      month = M and tonumber(M),
      day = d and tonumber(d),
      hour = h and tonumber(h),
      min = m and tonumber(m),
      sec = s and tonumber(s),
    }
  end
end

local function verifyDateFormat(val)
  if getDateTable(tostring(val)) then
    return val
  end
  return nil, 'date in format "[YYYY-MM-DD] hh:mm[:ss]"'
end

local function userListApply(path, sch, conf, prevConf)
  if not s.confCmp(conf, prevConf) then
    conf = conf or {}
    local etcPasswd, err = io.open"/etc/passwd"
    if not etcPasswd then
      return false, "Can't apply.\n"..err
    end
    local nsgUser = {}
    for line in etcPasswd:lines() do 
      local name, uid = line:match"^(.-):.-:(.-):.*$"
      if tonumber(uid) >= 500 then
        nsgUser[#nsgUser+1] = name
      end
    end
    for _, name in pairs(nsgUser) do
      if not conf[name] then
        os.execute("deluser "..name) 
      end
    end
  end
  return true
end

local function userApply(path, sch, conf, prevConf)
  if not s.confCmp(conf, prevConf) then
    local userName = path[#path]
    if userName ~= 'root' and userName ~= 'nsg' then
      os.execute(
        ('adduser %s -s /usr/sbin/nsgsh -D -H -h /home/user -g "NSG User" >/dev/null 2>&1'):
        format(userName)
      )
      C.addUser(userName, conf)
    end
  end
  return true
end

local function userRootNsgApply(path, sch, conf, prevConf)
  if not s.confCmp(conf, prevConf) then
    local userName = path[#path]
    C.updateRootNsgUser(userName, conf or {})
  end
  return true
end

local periodToSec = {
  hour = 60*60, 
  day = 60*60*24, 
  week = 60*60*24*7,
}
setmetatable(periodToSec, {__index = function(val)return val end})

local function restartNtpDaemon(path, sch, conf, prevConf)
  if not s.confCmp(conf, prevConf) then
    local ntpconf = s.addDefault(path, sch, conf, true) --clone conf with defaults
    if ntpconf.enable then
      local com = {
        "/usr/sbin/ntpclient",
        "-i", tostring(periodToSec[ntpconf.period]),
        "-h", ntpconf.host,
      }

      C.restartDaemon(path, com, {atExit = C.increasingSleep}) 
    else
      C.restartDaemon(path) 
    end
  end
  return true
end

local userSchemeScheme = s.newScheme{
  _H = "sysUserScheme",
  --[[
  _N = function(sch, conf)
    if conf and conf['@link'] then
      return {true, ['@link'] = true, ['@help'] = true}
    end
    return {}
  end,
  ]]
  ['@link'] = s.newScheme{
    _H = "sysUserSchemeLink",
    _T = "string",
    _V = s.stringVal,
    _DEP = 1,
  },
  ['@help'] = s.newScheme{
    _H = "sysUserSchemeHelp",
    _T = "string",
    _V = s.stringVal,
  },
}
userSchemeScheme._NAME = --userSchemeScheme

function(_, path)
  local link = C.getConfOrDefault(s.clonePath(path, -1, "@link"))
  return not link and userSchemeScheme
end

local userExcludeScheme = s.newScheme{
  _H = "sysUserExcludeList",
  _LIST = s.newScheme{
    _H = "sysUserExclude",
    _T = "string",
    _V = s.stringVal,
  },

}

local passwordScheme= s.newScheme{
  _H = "sysUserPassword",
  _X = function(path, sch, conf, sesNum)
    local userName = path[#path - 1]
    C.restartDaemon(path, {'passwd', userName}, 
      {sesNum = sesNum, noTimestamp = true, noRestart = true}) 
  end,
  _XI = true,
}

local userLocaleScheme = s.stringType{
  _H = "sysUserLocale",
}

local userOutputModeScheme = s.newScheme{
  _H = "sysUserOutputMode",
  _T = 'choice',
  _A = {"raw", "user-friendly"},
  _V = s.choiceVal,
  _D = "user-friendly",
}

local userScheme = s.newScheme{
  _H = "sysUser",
  _Z = userApply,
  permission = s.newScheme{
    _H = "sysUserPermission",
    _T = 'choice',
    _A = {
      {"rw", "Read/write"},
      {"ro", "Read only"},
    },
    _V = s.choiceVal,
    _D = "ro",
  },
  password = passwordScheme,
  scheme = userSchemeScheme{_P = -90},
  exclude = userExcludeScheme{_P = -100},
  locale = userLocaleScheme,
  ["output-mode"] = userOutputModeScheme,
}

local rootNsgUserScheme = s.newScheme{
  _H = "sysUser",
  _Z = userRootNsgApply,
  password = passwordScheme,
  locale = userLocaleScheme,
  ["output-mode"] = userOutputModeScheme,
}



systemScheme = s.newScheme{
  _Z = function() return true end,
  _H = "sys",
  hostname = s.newScheme{
    _H = "sysHostname",
    _T = "string",
    _V = s.stringVal,
    _Z = function(path, sch, conf, prevConf)
      if conf ~= prevConf and type(conf) == 'string' then
        os.execute(("hostname -- %q"):format(conf))
      end
      return true
    end
  },
  ["clock-set"] = s.newScheme{
    _H = "sysDate",
    _T = "string",
    _V = verifyDateFormat,
    _R = 'user', --TODO
    _X = function(path, sch, val)
      local usert = getDateTable(val)
      local t = os.date'*t'
      t.sec = 0
      for k, v in pairs(usert) do
        t[k] = v
      end
      os.execute(("date %02d%02d%02d%02d%04d.%02d >/dev/null 2>&1"):format(
          t.month, t.day, t.hour, t.min, t.year, t.sec
      ))
      return '\n'..s.readpopen("date 2>&1")
    end,
  },
  ["ppp-secrets"] = ppp.secretsScheme,
  configurations = nsgconfig.configScheme,
  software = soft.softwareScheme,
  users = s.newScheme{
    _H = "sysUsersList",
    _Z = userListApply,
    root = rootNsgUserScheme,
    nsg = rootNsgUserScheme,
    _NAME = userScheme,
  },
  sessions = s.newScheme{
    _H = "sysSessions",
    ["idle-time"] = s.newScheme{
      _H = "sysSessionsIdleTime",
      _T = "number",
      _A = {0, 60*60},
      _V = s.integerVal,
      _D = 60*5,

    },
    show = s.newScheme{
      _H = "sysSessionsShow",
      _X = function(path, sch, val, sesNum)
        return C.showSessions(sesNum)
      end,
      _R = 'user',
    },
    close = s.newScheme{
      _H = "sysSessionsClose",
      _T = "number",
      _A = {1, math.huge},
      _V = s.integerVal,
      _X = function(path, sch, val, sesNum)
        return C.closeOtherSessions(sesNum, val)
      end,
      _R = 'user',
    },
  },
  reboot = s.newScheme{
    _H = "sysReboot",
    _T = "string",
    _EX = { subtype = 'areyousure'},
    _V = s.areyousureVal,
    _A = {"System will be rebooted.\n"},
    _X = function(path, sch, val)
      if val == 'yes' then
        os.execute'reboot -d3&'
        return 'Bye-bye'
      else
        return nil
      end
    end
  },
  login = s.newScheme{
    _H = "sysLogin",
    _X = function(path, sch, conf, sesNum)
      C.restartDaemon(
        path, {'login'}, 
        { sesNum = sesNum, 
          noTimestamp = true, 
          noRestart = true,
          atKill = function(oper)
            if oper.daemon then
              oper.daemon:send("\nexit\n")
            end
          end,
        }
      ) 
    end,
    _XI = true,
  },

  watchdog = s.newScheme{
    _H = "sysWatchdog",
    _Z = function(path, sch, conf, prevConf)
      C.watchdog(conf and conf.enable)
      return true
    end,
    enable = s.booleanType{_H = "sysWatchdogEnable"},
    show = s.newScheme{
      _H = "sysWatchdogShow",
      _X = function(path, sch, val, sesNum)
        return C.watchdog'show'
      end,
      _R = 'user',
    },
  },

  ntp = s.newScheme{
    _H = "sysNtp",
    _Z = restartNtpDaemon,

    enable = s.booleanType{_H = "sysNtpEnable"},
    host = s.stringType{
      _H = "sysNtpHost",
      _D = "ru.pool.ntp.org"
    },
    period = s.newScheme{
      _T = 'string',
      _H = "sysNtpPeriod",
      _V = function(val)
        if (
          val == 'hour' or
          val == 'day' or
          val == 'week' or
          s.numberVal(val, 30, 604800)
        ) then
          return val
        end
        return nil, '"hour", "day", "week" or number of seconds in range 30..604800'
      end,
      _D = 'day',
    },
    ["log"] =  s.newScheme{
      _H = "sysNtpLog",
      _X = function(path)
        path = s.clonePath(path, -1)
        return C.getDaemonLog(path)
      end,
      _R = 'user',
    },
  },

  ["dns-client"] = dns.dnsClientScheme,

  show = s.newScheme{
    _H = "sysShow",
    date = s.newScheme{
      _H = "sysShowDate",
      _X = getOutOfUtil'date',
      _R = 'user',
    },
    uptime = s.newScheme{
      _H = "sysShowUptime",
      _X = getOutOfUtil'uptime',
      _R = 'user',
    },
    ps = s.newScheme{
      _H = "sysShowPs",
      _X = getOutOfUtil'ps',
      _R = 'user',
    },
    ["ip-route"] = s.newScheme{
      _H = "sysShowIproute",
      _X = getOutOfUtil'ip route show',
      _R = 'user',
    },
    dmesg = s.newScheme{
      _H = "sysShowDmesg",
      _X = getOutOfUtil'dmesg',
      _R = 'user',
    },
    who = s.newScheme{
      _H = "sysShowWho",
      _X = getOutOfUtil'who',
    },
    cpuinfo = s.newScheme{
      _H = "sysShowCpuinfo",
      _X = getOutOfUtil'cat /proc/cpuinfo',
      _R = 'user',
    },
    boardinfo = s.newScheme{
      _H = "sysShowBoardinfo",
      _X = function()
        return '\n'..(BI.info or "No Board Info")
      end,
      _R = 'user',
    },
    top = s.newScheme{
      _H = "sysShowTop",
      _X = function(path, sch, conf, sesNum)
        C.restartDaemon(path, {'top', '-b'}, 
          {sesNum = sesNum, noTimestamp = true, noRestart = true}) 
      end,
      _XI = true,
      _R = 'user',
    },
    version = s.newScheme{
      _H = "sysShowversion",
      _X = function()
        return C.getVersion()
      end,
      _R = 'user',
    },
    ["postmortem-dump"] = s.newScheme{
      _H = "sysShowPostmortemdump",
      _X = getOutOfUtil'cat /etc/postmortem.dump',
      _R = 'user',
    },
  },

  ['get-iface-name'] = function(_, path)
    return (C.session and C.session.script) and 
    s.stringType{
      _X = function(path, sch, val)
        return iface_mod.getIfaceName(val)
      end,
      _R = 'user',
      _P = -90,
    }
  end,

  ["log"] = s.newScheme{
    _H = "sysLog",
    _Z = function(path, sch, conf, prevConf)
      if not s.confCmp(conf, prevConf) then
        local conf = s.addDefault(path, sch, conf, true)
        C.setLogLevel(conf and conf.level)
        C.setLogMaxFilesize(conf and conf["max-filesize"])
      end
      return true
    end,
    filename = s.constantType {
      _H = "sysLogFile",
      _D = "/var/log/nsgconfd.log",
    },
    ["max-filesize"] = s.integerType{
      _H = "sysLogMaxfilesize",
      _A = {1024, 1024 * 1024},
      _D = 1024 * 100,
    },
    ["level"] = s.newScheme{
      _H = "sysLogLevel",
      _T = 'choice',
      _A = {"debug", "info", "warn", "error", "fatal"},
      _V = s.choiceVal,
      _D = "info",
    },
    show = s.newScheme{
      _H = "sysLogShow",
      _X = function(path, sch, val, sesNum)
        return('\n'..s.readpopen("tail -n100 "..
            C.getConfOrDefault(s.clonePath(path, -1, "filename")).." 2>&1"))
      end,
      _R = 'user',
    },
  },


}
