module(..., package.seeall)
local s = require'scheme'
local lfs = require'lfs'
local uita = require'uita'

local ip = "/usr/sbin/ip"

local uitcpRestartTimestamp

local uitcpConfigFilename
local uitcpConfigPath
local uitcpConfigScheme

local function updateUitcpConfigFile()
  if (uitcpConfigFilename and 
    lfs.attributes(uitcpConfigFilename, 'mode') == 'file'
  ) then

    local conf = s.addDefault(uitcpConfigPath, uitcpConfigScheme, 
      C.getConfByPath(uitcpConfigPath), true
    )
    local f = io.open(uitcpConfigFilename, 'w')
    if f then
      f:write("return", s.ser(conf.configure))
      f:close()
    end
  end
end

local function applyUitcp(func, name)
  return function(path, sch, conf, prevConf)
    if (
      (
        not uitcpRestartTimestamp or
        os.time() - uitcpRestartTimestamp > 0
      ) and
      not s.confCmp(conf, prevConf) 
    ) then
      updateUitcpConfigFile()
      uitcpRestartTimestamp = nil
      local conf = s.addDefault(path, sch, conf, true) --clone conf with defaults
      local res, err
      if name then
        res, err = uita[func](path[#path], conf)
      else
        res, err = uita[func](conf)
      end
      if not res and err == "connection refused" then
        res = true
        err = nil
      end
      return res, err
    end
    return true
  end
end

local uitcpBinFile = {
  '/etc/uitcp/bin/uitcp',
  '/usr/sbin/uitcp'
}

local function restartUitcp(path, sch, conf, prevConf)
  local conf = s.addDefault(path, sch, conf, true) --clone conf with defaults
  uitcpRestartTimestamp = os.time()
  if conf.enable then
    uitcpConfigPath = path
    uitcpConfigScheme = sch
    uitcpConfigFilename = os.tmpname()
    local f = io.open(uitcpConfigFilename, 'w')
    if f then
      f:write("return", s.ser(conf.configure))
      f:close()
    end

    local uitcpUtil
    for i = 1, #uitcpBinFile do
      local f = io.open(uitcpBinFile[i], 'r')
      if f then
        f:close()
        uitcpUtil = uitcpBinFile[i]
        break
      end
    end

    if uitcpUtil then
      C.restartDaemon(
        path, 
        {uitcpUtil, '-n', uitcpConfigFilename }, 
        { 
          noTimestamp = true,
          tmpFileList = {uitcpConfigFilename}, 
          atExit = C.increasingSleep
        }
      )
    else
      C.restartDaemon(path)
    end
    return true
  end
  C.restartDaemon(path)
  return true
end


local tcpKeepaliveIntervalScheme = s.newScheme{
  _T = "number", _A = {1, 0x7fffffff}, _V = s.integerVal, 
  _H = "uitcpTcpKeepaliveInterval",
}
local tcpKeepaliveRetryScheme = s.newScheme{
  _T = "number", _A = {1, 0x7fffffff}, _V = s.integerVal, 
  _H = "uitcpTcpKeepaliveRetry",
}

local natRule = s.newScheme {
  _H = "uitcpNatX",
  inId = s.stringType{ 
    _H = "uitcpNatInId",
    _P = 100,
  },
  inSrcAddr = s.ipAddressType{_H = "uitcpNatInSrcAddr"},
  inDstPort = s.tcpPortType{_H = "uitcpNatInDstPort"},
  outSrcAddr = s.ipAddressType{_H = "uitcpNatOutSrcAddr"},
  outDstAddr = s.ipAddressType{_H = "uitcpNatOutDstAddr"},
  outDstPort = s.tcpPortType{_H = "uitcpNatOutDstPort"},
  tcpKeepaliveInterval = tcpKeepaliveIntervalScheme,
  tcpKeepaliveRetry = tcpKeepaliveRetryScheme,
}

local natList = s.newScheme{
  _H = "uitcpNat",
  _LIST = natRule,
}

local sslScheme = s.newScheme{
  _H = "uitcpSsl",
  path = s.newScheme{
    _T = "string", 
    _H = "uitcpSslPath",
    _V = s.filenameVal,
  },
  protocol = s.newScheme{
    _T = "choice",
    _A = {"tlsv1", "sslv3", "sslv23"}, 
    _V = s.choiceVal, 
    _H = "uitcpSslProtocol",
    _D = 'sslv3',
  },
  key = s.newScheme{
    _T = "string", 
    _H = "uitcpSslKey",
    _V = s.filenameVal,
  },
  password = s.newScheme{
    _T = "string",
    _H = "uitcpSslPassword",
    _V = s.stringVal,
  },
  certificate = s.newScheme{
    _T = "string", 
    _H = "uitcpSslCertificate",
    _V = s.filenameVal,
  },
  cafile = s.newScheme{
    _T = "string", 
    _H = "uitcpSslCafile",
    _V = s.filenameVal,
  },
  capath = s.newScheme{
    _T = "string", 
    _H = "uitcpSslCapath",
    _V = s.filenameVal,
  },
  verify = s.newScheme{
    _H = "uitcpSslVerify",
    none = s.booleanType{_H = "uitcpSslVerifyNone"},
    peer = s.booleanType{_H = "uitcpSslVerifyPeer", _D = true},
    client_once = s.booleanType{_H = "uitcpSslVerifyClientOnce"},
    fail_if_no_peer_cert = s.booleanType{
      _H = "uitcpSslVerifyFailifnopeercert", 
      _D = true
    },
  },
  options = s.newScheme{
    _H = "uitcpSslOptions",
    ["all"] = s.booleanType{_H = "uitcpSslOptionAll", _D = true},
    ["cipher_server_preference"] = s.booleanType{_H = "uitcpSslOptionCipherserverpreference"},
    ["dont_insert_empty_fragments"] = s.booleanType{_H = "uitcpSslOptionDontinsertemptyfragments"},
    ["ephemeral_rsa"] = s.booleanType{_H = "uitcpSslOptionEphemeralrsa"},
    ["netscape_ca_dn_bug"] = s.booleanType{_H = "uitcpSslOptionNetscapecadnbug"},
    ["netscape_challenge_bug"] = s.booleanType{_H = "uitcpSslOptionNetscapechallengebug"},
    ["microsoft_big_sslv3_buffer"] = s.booleanType{_H = "uitcpSslOptionMicrosoftbigsslv3buffer"},
    ["microsoft_sess_id_bug"] = s.booleanType{_H = "uitcpSslOptionMicrosoftsessidbug"},
    ["msie_sslv2_rsa_padding"] = s.booleanType{_H = "uitcpSslOptionMsiesslv2rsapadding"},
    ["netscape_demo_cipher_change_bug"] = s.booleanType{_H = "uitcpSslOptionNetscapedemocipherchangebug"},
    ["netscape_reuse_cipher_change_bug"] = s.booleanType{_H = "uitcpSslOptionNetscapereusecipherchangebug"},
    ["no_session_resumption_on_renegotiation"] = s.booleanType{_H = "uitcpSslOptionNosessionresumptiononrenegotiation"}, 
    ["no_sslv2"] = s.booleanType{_H = "uitcpSslOptionNosslv2", _D = true},
    ["no_sslv3"] = s.booleanType{_H = "uitcpSslOptionNosslv3"},
    ["no_tlsv1"] = s.booleanType{_H = "uitcpSslOptionNotlsv1"},
    ["pkcs1_check_1"] = s.booleanType{_H = "uitcpSslOptionPkcs1check1"},
    ["pkcs1_check_2"] = s.booleanType{_H = "uitcpSslOptionPkcs1check2"},
    ["single_dh_use"] = s.booleanType{_H = "uitcpSslOptionSingledhuse"},
    ["ssleay_080_client_dh_bug"] = s.booleanType{_H = "uitcpSslOptionSsleay080clientdhbug"},
    ["sslref2_reuse_cert_type_bug"] = s.booleanType{_H = "uitcpSslOptionSslref2reusecerttypebug"},
    ["tls_block_padding_bug"] = s.booleanType{_H = "uitcpSslOptionTlsblockpaddingbug"},
    ["tls_d5_bug"] = s.booleanType{_H = "uitcpSslOptionTlsd5bug"},
    ["tls_rollback_bug"] = s.booleanType{_H = "uitcpSslOptionTlsrollbackbug"},
    --OpenSSL 0.9.8 only
    ["cookie_exchange"] = s.booleanType{_H = "uitcpSslOptionCookieexchange"},
    ["no_query_mtu"] = s.booleanType{_H = "uitcpSslOptionNoquerymtu"},
    ["single_ecdh_use"] = s.booleanType{_H = "uitcpSslOptionSingleecdhuse"},
  },
  ciphers = s.newScheme{
    _T = "string", 
    _H = "uitcpSslCiphers"
  },
  depth = s.newScheme{
    _T = "integer", _A = {1}, _V = s.integerVal, 
    _H = "uitcpSslDepth",
  },
  certcheck =  s.newScheme{
    _T = "boolean", _V = s.booleanVal,
    _H = "uitcpSslCertcheck",
  },
}

local listenerScheme = s.newScheme{
  _H = "uitcpListenerX",
  id = s.stringType{ 
    _H = "uitcpListenerId",
    _P = 100,
  },
  host = s.ipAddressOrAllType{_H = "uitcpListenerHost"},
  port = s.tcpPortType{_H = "uitcpListenerPort"},
  tcpKeepaliveInterval = tcpKeepaliveIntervalScheme,
  tcpKeepaliveRetry = tcpKeepaliveRetryScheme,
}

local listOfListenerScheme = s.newScheme{
  _H = "uitcpListener",
  _LIST = listenerScheme,
}

local ifAddressScheme = s.newScheme{
  _H = "uitcpIfAddress",
  prefix = s.ipPrefixType{_H = "uitcpIfAddressPrefix", _P = 100},
  peer = s.ipAddressType{_H = "uitcpIfAddressPeer", _P = 90},
  broadcast = s.ipAddressType{_H = "uitcpIfAddressBroadcast", _P = 80},
  anycast = s.ipAddressType{_H = "uitcpIfAddressAnycast", _P = 70},
}

local ifAddressListScheme = ifAddressScheme{
  _H = "uitcpSocketNetIfAddress",
  _LIST = ifAddressScheme{
    _H = "uitcpSocketNetIfAddressX"
  }
}

local socketScheme = s.newScheme{
  _H = "uitcpSocket",
  _N =  function(sch, conf)
    conf = conf or {}
    if conf.type == 'net' then
      return {nil, raw = true, udp = true}
    elseif conf.type == 'raw' then
      return {nil, net = true, udp = true}
    else 
      return {nil, net = true, raw = true}
    end
  end,
  disable =  s.newScheme{
    _T = "boolean", _V = s.booleanVal,
    _H = "uitcpSocketDisable",
    _P = 100,
  },
  ["type"] = s.newScheme{
    _T = 'choice',
    _A = {'raw', 'udp', 'net'},
    _V = s.choiceVal,
    _H = "uitcpSocketType",
    _D = 'udp',
    _P = 50,
    _DEP = 1,
  },
  udp = s.newScheme{
    _H = "uitcpSocketUdp",
    peerAddress = s.ipAddressType{_D = '127.0.0.1', _H = "uitcpSocketUdpPeerAddress"},
    peerPort = s.tcpPortType{_D = 0, _H = "uitcpSocketUdPeerPort"},
    socketAddress = s.ipAddressOrAllType{_D = '*', _H = "uitcpSocketUdpSocketAddress"},
    socketPort = s.tcpPortType{_D = 0, _H = "uitcpSocketUdpSocketPort"},
  },
  raw = s.newScheme{
    _H = "uitcpSocketRaw",
    protocol = s.newScheme{
      _T = "integer",
      _A = {0,0xffff}, 
      _V = integerVal, 
      _H = "uitcpSocketRawProtocol",
    },
    ipHdrIncl = s.newScheme{
      _T = "boolean", _V = s.booleanVal,
      _H = "uitcpSocketRawIpHdrIncl",
    },
    peerAddress = s.ipAddressType{_D = '127.0.0.1', _H = "uitcpSocketRawPeerAddress"},
    socketAddress = s.ipAddressOrAllType{_D = '*', _H = "uitcpSocketRawSocketAddress"},
  },
  net = s.newScheme{
    _H = "uitcpSocketNet",
    ifName = s.stringType{ 
      _H = "uitcpSocketNetIfName",
    },
    ifAddress = ifAddressListScheme,
  }
}

local function multiTunnelSocketNameValidator(val)
  return s.anyVal(val)
end

local multiTunnelSocketScheme = socketScheme{
  qdisc = s.newScheme{
    _T = 'choice',
    _A = {'roundRobin', 'fullWindow', 'weighted'},
    _V = s.choiceVal,
    _H = "uitcpMultiTunnelSocketQdisc",
    _D = 'roundRobin',
  },
}

local multiTunnelSocketPointerScheme = s.newScheme{
  _H = "uitcpMultiTunnelSocketPointer",
  _T = "string",
  _V = multiTunnelSocketNameValidator,
}

local setOfSocketsScheme = s.newScheme{
  _H = "uitcpSockets",
  _NAME = socketScheme,
  _NAME_TYPE =  function(val)
    return s.integerVal(val, 1, 64)
  end,
}

local function secretsVal(val)
  if val == "default" or type(val) == 'table' then
    return val
  end
  return nil, [[List of file names, e.g. {c="cert",k="key",r="root"}'
or 'default'
]]
end

local clientToolsScheme = s.newScheme{
  _H = "uitcpClientsTools",
  secrets = s.newScheme{
    _H = "uitcpClientsToolsSecret",
    _T = "string", 
    _V = secretsVal,
    _X =  function(path, sch, v) 
      local t, str = uita.secrets(path[#path-2], v)
      return t and s.toStr(t) or str
    end,
  },  
}

local clientScheme = s.newScheme{
  _Z = applyUitcp('restartTun', true),
  _H = "uitcpClientsX",
  description = s.stringType{_H = "uitcpClientsDescription"},
  localListener = listOfListenerScheme,
  socket = setOfSocketsScheme,
  nat = natList,
  ssl = sslScheme{
    disable =  s.newScheme{
      _T = "boolean", _V = s.booleanVal,
      _H = "uitcpSslDisable",
    },
  },
  multiTunnelSocket = multiTunnelSocketPointerScheme{_P = -100},
  tools = clientToolsScheme{_P = -200},
}

local serversListScheme = s.newScheme{
  _H = "uitcpServersList",
  _NAME = s.ipAddressAndPortType{_D = "127.0.0.1:50005", _H = "uitcpServerAddr"},
  _NAME_TYPE =  function(val)
    return s.integerVal(val, 1, 64)
  end,
}

local restartTmoScheme =  s.newScheme{
  _T = "number", _A = {1, 60*60*24}, _V = s.numberVal, 
  _H = "uitcpTunnelLinksRestartTimeout",
}

local linksScheme = s.newScheme{
  _H = "uitcpTunnelLinksX",
  disable =  s.newScheme{
    _T = "boolean", _V = s.booleanVal,
    _H = "uitcpTunnelLinksDisable",
  },
  dev = s.stringType{_H = "uitcpTunnelLinksDev"},
  description = s.stringType{_H = "uitcpTunnelLinksdescription"},
  servers = serversListScheme,
  gw = s.ipAddressType{_H = "uitcpTunnelLinksGw"},
  restart = s.newScheme{
    _H = "uitcpTunnelLinksRestart",
    _LIST = s.newScheme{
      _H = "uitcpTunnelLinksRestartX",
      description = s.stringType{_H = "uitcpTunnelLinksRestartDescription"},
      script = s.stringType{_H = "uitcpTunnelLinksRestartScripts"},
      servers = serversListScheme,
      timeout = restartTmoScheme,
    },
    timeout = restartTmoScheme{_D = 120},
    primary = function(_, path)
      return path[#path - 2] == 1 and
      s.integerType{
        _H = "uitcpTunnelLinksPrimary",
        _A = {1, math.huge}
      }
      or false
    end,
  },
  csqRequest = s.newScheme{
    _H = "uitcpTunnelLinkssqRequest",
    ttydev = s.stringType{_H = "uitcpTunnelLinkssqRequestTtydev"},
    interval =  s.newScheme{
      _T = "number", _A = {1, 60*60*24}, _V = s.numberVal, 
      _H = "uitcpTunnelLinkssqRequestInterval",
      _D = 60,
    },
  }
}

local linksList = s.newScheme{
  _H = "uitcpTunnelLinks",
  _LIST = linksScheme,

  priority = s.integerType{
    _H = "uitcpTunnelLinksPriority",
    _A = {0, math.huge},
  },
  ["idle-time"] = function(_, path)
    return C.getConfOrDefault(s.clonePath(path, -1, "priority")) and
    s.integerType{
      _H = "uitcpTunnelLinksIdeletime",
      _A = {0, math.huge},
    }
  end,
}

local tunnelScheme = s.newScheme{
  _Z = applyUitcp('restartMainTun'),
  _H = "uitcpTunnel",
  disable =  s.newScheme{
    _T = "boolean", _V = s.booleanVal,
    _H = "uitcpTunnelDisable",
  },
  description = s.stringType{_H = "uitcpTunnelDescription"},
  servers = serversListScheme{_P = 51},
  numOfServers = s.newScheme{
    _T = "integer", _A = {1, 8}, _V = s.integerVal, 
    _H = "uitcpTunnelNumOfServers",
    _D = 1,
    _P = 50,
  },
  window = s.newScheme{
    _T = "integer", _A = {1, 128}, _V = s.integerVal, 
    _H = "uitcpTunnelWindow",
    _D = 8,
    _P = 10,
  },
  keepalive = s.newScheme{
    _T = "number", _A = {1, 60*59+0.99}, _V = s.numberVal, 
    _H = "uitcpTunnelKeepalive",
    _D = 5,
    _P = 9,
  },
  attempt = s.newScheme{
    _T = "integer", _A = {1, 128}, _V = s.integerVal, 
    _H = "uitcpTunnelFttempt",
    _D = 3,
    _P = 8,
  },
  connectTmo = s.newScheme{
    _T = "number", _A = {1, 60*60*24}, _V = s.numberVal, 
    _H = "uitcpTunnelConnectTmo",
    _D = 20,
    _P = 7,
  },
  connectAttempt = s.newScheme{
    _T = "integer", _A = {1, 128}, _V = s.integerVal, 
    _H = "uitcpTunnelConnectAttempt",
    _D = 10,
    _P = 6
  },

  links = linksList,
  localListener = listOfListenerScheme,
  socket = setOfSocketsScheme{
    _P = -10
  },
  nat = natList,
  ssl = sslScheme{
    disable =  s.newScheme{
      _T = "boolean", _V = s.booleanVal,
      _H = "uitcpSslDisable",
    },
  },
  multiTunnelSocket = multiTunnelSocketPointerScheme{_p = -100},
  --linksPriority
}

local logScheme = s.newScheme{
  _Z = applyUitcp('restartLog'),
  _H = "uitcpLog",
  level = s.newScheme{
    _T = "choice", 
    _A = {'debug', 'info', 'warn', 'error', 'fatal'}, 
    _V = s.choiceVal, 
    _H = "uitcpLogLevel",
    _D = 'info',
  },
  trace = s.newScheme{
    _T = "integer", 
    _A = {0, 1024}, 
    _V = s.integerVal, 
    _H = "uitcpLogTrace",
  },
  logFile = s.newScheme{ 
    _T = "string", _V = s.filenameVal,
    _H = "uitcpLogLogFile",
    _D = "/var/log/uitcp",
  },
  logFileMaxSize = s.newScheme{
    _T = "integer", 
    _A = {1024, 1024*1024*1024}, 
    _V = s.integerVal, 
    _H = "uitcpLogLogFileMaxSize",
    _D = 100*1024,
  },
  sql = s.newScheme{
    _H = "uitcpLogSql",
    database = s.newScheme{
      _T = "choice", 
      _A = {'PostgreSQL', 'ODBC', 'MySQL', 'SQLite', 'Oracle', 'ADO'}, 
      _V = s.choiceVal, 
      _H = "uitcpLogSqlDatabase",
    },
    sourcename = s.newScheme{
      _T = 'string',
      _V = s.stringVal,
      _H = "uitcpLogSqlSourcename"
    },
    username = s.newScheme{
      _T = 'string',
      _V = s.stringVal,
      _H = "uitcpLogSqlUsername"
    },
    password = s.newScheme{
      _T = 'string',
      _V = s.stringVal,
      _H = "uitcpLogSqlPassword"
    },
    hostname = s.newScheme{
      _T = 'string',
      _V = s.stringVal,
      _H = "uitcpLogSqlHostname"
    },
    port = s.tcpPortType{_H = "uitcpLogSqlPort"},
    tablename = s.newScheme{
      _T = 'string',
      _V = s.stringVal,
      _H = "uitcpLogSqlTablename"
    },
    fields = s.newScheme{
      _H = "uitcpLogSqlFields",
      date = s.newScheme{
        _T = 'string',
        _V = s.stringVal,
        _H = "uitcpLogSqlFieldsDate"
      },
      level = s.newScheme{
        _T = 'string',
        _V = s.stringVal,
        _H = "uitcpLogSqlFieldsLevel"
      },
      hostname = s.newScheme{
        _T = 'string',
        _V = s.stringVal,
        _H = "uitcpLogSqlFieldsHostname"
      },
      tunnelname = s.newScheme{
        _T = 'string',
        _V = s.stringVal,
        _H = "uitcpLogSqlFieldsTunnelname"
      },
      message = s.newScheme{
        _T = 'string',
        _V = s.stringVal,
        _H = "uitcpLogSqlFieldsMessage"
      },
    }
  }
}

local sysScheme = s.newScheme{
  _H = "uitcpSys",
  _Z = applyUitcp('restartUitcpSys'),
  pluginFolder = s.newScheme{ 
    _T = "string", _V = s.filenameVal,
    _H = "uitcpSysPluginFolder",
    _D = '/etc/uitcp/plugins',
  },
}

local showScheme = s.newScheme{
  _H = "uitcpShow",
  _HTML = "uitcpweb",
  version = s.newScheme{
    _H = "uitcpShowVersion",
    _X =  function(path, sch, v) 
      local t, str = uita.ver()
      return t and s.toStr(t) or str
    end,
  },
  stat = s.newScheme{
    _H = "uitcpShowStat",
    _T = 'string',
    _V = s.stringVal,
    _X =  function(path, sch, v) 
      local t, str = uita.stat(v)
      return t and s.toStr(t) or str
    end,
  },
  connnections = s.newScheme{
    _H = "uitcpShowConnnections",
    _T = 'string',
    --_V = s.stringVal,
    _X =  function(path, sch, v) 
      local t, str = uita.con(v)
      return t and s.toStr(t) or str
    end,
  },
  certificate = s.newScheme{
    _H = "uitcpShowCertificate",
    _X =  function(path, sch, v) 
      local t, str = uita.certs(v)
      if t then
        local list = {}
        for k, v in pairs(t) do
          list[#list+1] = {k, v}
        end
        table.sort(list, function(a, b)return a[2]<b[2]end)
        local out = {}
        local curT = os.time()
        for _, v in ipairs(list) do
          out[#out+1] = ("%-40s expires %s in %4d days"):format(
            v[1]:sub(-40), os.date('%F', v[2]), 
            (1/(60*60*24)) * (v[2] - curT)
          )
        end
        return table.concat(out, '\n')
      else
        return str
      end
    end,
  },
}

local uitcpConfigureScheme = s.newScheme{
  _H = "uitcpConf",
  _Z =  function(path, sch, conf) 
    return true
  end,
  _N =  function(sch, conf)
    local mode = conf and conf.mode or sch.mode._D
    if mode == 'server' then
      return {nil, tunnel = true, tunnels = true}
    else 
      return {nil, clients = true, tunnelListener = true}
    end
  end,
  _V =  function(conf)
    if type(conf) ~= 'table' then
      return nil, 'For mode = "client" parameter "tunnel" is mandatory'
    end
    if conf.mode == 'client' then
      if not conf.tunnel then
        return nil, "Parameter 'tunnel' is mandatory"
      end
    else
    end
    return true
  end,

  mode = s.newScheme{
    _Z = applyUitcp('restartMode'),
    _T = 'choice',
    _A = {'client', 'server',},
    _V = s.choiceVal,
    _H = "uitcpMode",
    _D = 'client',
    _P = 100,
    _DEP = 1,
  },

  nat = natList{
    _Z = applyUitcp('restartNAT'),
    _H = "uitcpNat"
  },

  ssl = sslScheme{
    _Z = applyUitcp('restartSSL'),
    _H = "uitcpSslDefault"
  },

  tunnelListener = s.newScheme{
    _Z = applyUitcp('restartTL'),
    _H = "uitcpTunnelListener",
    host = s.ipAddressOrAllType{_H = "uitcpTunnelListenerHost", _D = '*'},
    port = s.tcpPortType{_H = "uitcpTunnelListenerPort", _D = 50005},
    _LIST = listenerScheme{_H = "uitcpTunnelListenerExtra"},
  },
  clients = s.newScheme{
    _Z = applyUitcp('restartClients'),
    _H = "uitcpClients",
    _NAME = clientScheme,
  },
  tunnel = tunnelScheme{
    _V =  function(cnf)
      if type(cnf) == 'table' and cnf.name and cnf.links then
        return cnf
      end
      return nil, 'name and links are mandatory'
    end,

    name = s.stringType{ 
      _H = "uitcpTunnelName",
      _P = 100,
    },
  },
  tunnels = s.newScheme{
    _Z = applyUitcp('restartTunnels'),
    _H = "uitcpTunnels",
    _NAME = tunnelScheme{
      _Z = applyUitcp('restartClientTun', true),
    },
  },
  log = logScheme,
  sys = sysScheme,
  multiTunnelSocket = s.newScheme{
    _P = -100,
    _H = "uitcpMultiTunnelSocket",
    _NAME = multiTunnelSocketScheme,
    _NAME_TYPE =  multiTunnelSocketNameValidator,
    _Z = applyUitcp('restartMultiTunnelSocket'),
  },
  show = showScheme{_P = -180},
  SHUTDOWN = s.newScheme{
    _P = -200,
    _H = "uitcpShutdown",
    _T = "boolean", _V = s.booleanVal,
    _X =  function(path, sch, v) 
      local t, str = uita.shutdown(v)
      return t and s.toStr(t) or str
    end,
  },
}


uitcpScheme = s.newScheme{
  _H = "uitcp",
  _Z = restartUitcp,
  enable = s.booleanType{_H = "uitcpEnable"},
  configure = uitcpConfigureScheme,
  ["log"] =  s.newScheme{
    _H = "uitcpLogX",
    _X = function(path)
      return C.getDaemonLog(s.clonePath(path, -1))
    end,
    _R = 'user',
  },
}
