--------------------------------------------------------------------------------
-- Module scheme.
-- reserved names:
-- _A - arguments for verification.
-- _D - default value.
-- _H - helpName in `help` module
-- _N - needless parameter .
-- _T - type of parameter.
-- _R - rights: 'user', 'admin', 'root'
-- _V - verification function.
-- _X - exec function.
-- _Z - apply function.
-- _HTML - string with external page generator name (*.lua)
-- _EX - table with extended flags fitures and so on
--
-- _DEP - INTEGER levels up for rebuild (1 for rebuild siblings).
-- _LIST -
-- _NAME -
-- _NAME_TYPE -
-- _methods - table of methods

module(..., package.seeall)

-- Common functions

local function pathConcat (...)
  local r = {}
  for _, l in ipairs ({...}) do
    for _, v in ipairs (l) do
      table.insert (r, v)
    end
  end
  return r
end
local function pathConcat2 (base, path, prefix)
  if not base then return end
  local r = {}
  for i = 1, #base do
    r[i] = base[i]
  end
  for i = (prefix or 0) + 1, #path do
    r[#r+1] = path[i]
  end
  return r
end

function clonePath(path, n, ...)
  if type(path) ~= 'table' then return nil end
  local newPath = {}
  n = n or #path
  if n < 0 then n = #path + n end
  for i = 1, n do
    newPath[i] = path[i]
  end
  for _, v in ipairs{...} do
    newPath[#newPath+1] = v
  end
  return newPath
end

function deepcopy(object)
  local lookup_table = {}
  local function _copy(object)
    if type(object) ~= "table" then
      return object
    elseif lookup_table[object] then
      return lookup_table[object]
    end
    local new_table = {}
    lookup_table[object] = new_table
    for index, value in pairs(object) do
      new_table[_copy(index)] = _copy(value)
    end
    return setmetatable(new_table, getmetatable(object))
  end
  return _copy(object)
end

function readpopen(prog)
  local f = io.popen(prog)
  local s
  if f then
    s = f:read("*a")
    f:close()
  end
  return s
end


local strf = string.format
function tostr(s)
  local out
  if type(s) == "string" then
    out = strf('%q', s):gsub('\\\n', '\\n')
  else
    out = tostring(s)
  end
  return out
end

local nsg = require'nsg'
local help = require'nsgconfhelp'

ser = nsg.serialize

function toStr(s)
  local out
  if s == nil then
    out = ''
  elseif type(s) == 'table' then
    out = ser(s, 0)..'\n'
  else
    out = tostring(s)..'\n'
  end
  return out
end

local mainPairs do
  local function mainPairsAux(t, keys)
    for k, v in pairs(t) do
      if not (type(k) == "string" and k:sub(1,1) == '_') then
        if v and not keys[k] then
          keys[k] = v
          --coroutine.yield(k, v)
        end
      end
    end
    local m = getmetatable(t)
    if m and type(m.__index) == 'table' then
      mainPairsAux(m.__index, keys)
    end
  end
  
  local function mainPairsIterator(t)
    local res = {}
    mainPairsAux(t, res)
    for k, v in pairs(res) do
      coroutine.yield(k, v)
    end
  end

  local curSchemeNode, curPath
  local function compareObject(a,b)
    local aSch = curSchemeNode[a](nil, clonePath(curPath, nil, a))
    local bSch = curSchemeNode[b](nil, clonePath(curPath, nil, b))
    local prioA = type(aSch)=="table" and aSch._P or 0
    local prioB = type(bSch)=="table" and bSch._P or 0
    if prioA ~= prioB then
      return prioA > prioB
    end
    local typeA = type(a)
    local typeB = type(b)
    if typeA == typeB then
      if typeA == "number" or typeA == "string" then
        return a < b
      else
        return false
      end
    end
    local t = {
      ["nil"] = 1,
      boolean = 2,
      number = 3,
      string = 4,
      table = 5,
      ["function"] = 6,
      thread = 7,
      userdata = 8,
    }
    return (t[typeA] or 10) < (t[typeB] or 10)
  end

  local function mainPairsSorted(t, path)
    local a = {}
    for n in t:_pairs() do a[#a + 1] = n end
    curSchemeNode = t
    curPath = path
    table.sort(a, compareObject)
    local i = 0
    return function()
      i = i + 1
      return a[i], t[a[i]]
    end
  end

  function mainPairs(t, sortPath)
    if sortPath then
      return mainPairsSorted(t, sortPath)
    else
      return coroutine.wrap(function() mainPairsIterator(t) end)
    end
  end
end

local function aux(self, at)

  if at == nil then
    return self
  elseif type(at) == 'string' or type(at) == 'number' then
    return self[at]
  elseif type(at) == 'table' then
    local mt = {__index = self, __call = aux}
    setmetatable(at, mt)
    return at
  end
  return self
end

local methods = {
  text = require'textmethods',
  html = require'htmlmethods',
}

--local confPairs
function newScheme(t)
  t.__call = aux
  t._pairs = mainPairs
  --t._confpairs = confPairs
  t._methods = methods
  setmetatable(t, t)
  return t
end

local function eachNode(absPath, scheme, conf, path, onlyValid, sorted, pathPrefix)
  if not absPath and type(scheme._LINK) == 'table' then
    absPath = scheme._LINK
    pathPrefix = #path
  end
  path = path or {}
  --local n = #path + 1
  if type(scheme) ~= "table" then
    if not onlyValid then
      coroutine.yield(path or true, false, conf)
    end
    return
  end
  coroutine.yield(path or true, scheme, conf)
  if not scheme._T then
    local need = (scheme._N and conf) and scheme._N(scheme, conf, path or {}) or {}
    if type(conf) ~= 'table' then
      for k, v in scheme:_pairs(sorted and (path or {})) do
        local path = clonePath(path, nil, k)
        if not onlyValid or need[k] == need[1] then
          eachNode(
            absPath, v(nil, pathConcat2(absPath, path, pathPrefix)), nil, 
            path, onlyValid, sorted, pathPrefix
          )
        end
      end
    else -- conf is table
      local confKeys = {}
      for k, v in scheme:_pairs(sorted and (path or {})) do
        local path = clonePath(path, nil, k)
        confKeys[k] = true
        if not onlyValid or need[k] == need[1] then
          eachNode(
            absPath, v(nil, pathConcat2(absPath, path, pathPrefix)), conf[k], 
            path, onlyValid, sorted, pathPrefix
          )
        end
      end

      local confKeysList = {}
      for k, _ in pairs(conf) do
        if not confKeys[k] then
          confKeysList[#confKeysList+1] = k
        end
      end
      if sorted then table.sort(confKeysList, nsg.compareObject) end
      for _, k in ipairs(confKeysList) do
        local path = clonePath(path, nil, k)
        if not onlyValid or need[k] == need[1] then
          local descr = (type(k) == "number" and scheme._LIST) or
          ((scheme._NAME_TYPE or isDefaultNameType)(k) and
          scheme._NAME)
          if descr then
            eachNode(
              absPath, descr(nil, pathConcat2(absPath, path, pathPrefix)), conf[k], 
              path, onlyValid, sorted, pathPrefix
            )
          else
            if not onlyValid or type(k) == "string" and k:sub(1,1) == '_' then
              coroutine.yield(path or true, nil, conf[k])
            end
          end
        end
      end
    end
  else
  end
end

--- Iterator through all nodes of scheme and configuration tables
-- @param scheme scheme table
-- @param conf configuration table
-- @param onlyValid return only valid nodes
-- @return path, scheme, conf
function allNodes(absPath, scheme, conf, onlyValid, sorted)
  return coroutine.wrap(
    function() eachNode(absPath, scheme, conf, absPath and {} or nil, onlyValid, sorted) end
  )
end

do
  local function auxconfPairs(absPath, scheme, conf, onlyValid, sorted)
    for path, sch, cnf in allNodes(absPath, scheme, conf, onlyValid, sorted) do
      if cnf ~= nil then
        coroutine.yield(path, cnf)
      end
    end
  end

  function confPairs(path, scheme, conf, onlyValid, sorted)
    return coroutine.wrap(function() 
        auxconfPairs(path, scheme, conf, onlyValid, sorted) 
    end)
  end
end

--- Add value to table.
-- This function add value to table as t[k1][k2][k3] = val.
-- Keys k1, k2, k3,... may not exist in table(they will be added).
-- @param t target table
-- @param path list of keys such as {k1, k2, k3,...}
-- @param val value
function addValToTable(t, path, val)
  if #path == 0 or type(t) ~= 'table' then return val end
  local nt = t
  local key
  for _, v in ipairs(path) do
    if key  then
      if( not nt[key]) then
        nt[key] = {}
      end;
      nt = nt[key]
      if type(nt) ~= 'table' then return val end
    end
    key = v
  end
  nt[key] = val
  return t
end

--- Validate configuration concerning scheme.
-- @param scheme scheme table
-- @param conf configuration table
-- @param dflt if true then bad value will replaced with default ones
-- @return table with error mesages
function confValidate(absPath, scheme, conf, dflt)
  local errtab = {}
  local res = true
  for path, sch, cnf in allNodes(absPath, scheme, conf) do
    if cnf ~= nil then
      if sch then
        if sch._N then
          local need = sch._N(sch, cnf, path or {})
          for k, _ in sch:_pairs() do
            if cnf[k] and need[k] ~= need[1] then 
              if dflt then
                cnf[k] = nil
              else
                path[#path+1] = k
                addValToTable(errtab, path, "Needless parameter")
                path[#path] = nil
                res = nil
              end
            end
          end
        end
        if  not sch._T and type(cnf) ~= 'table' then -- conf must be table
          if dflt then
            addValToTable(conf, path, nil)
          else
            if #path == 0 then path[1] = '_ERROR' end
            addValToTable(errtab, path, "Must be table")
            res = nil
          end
        elseif sch._V then
          local r, e = sch._V(cnf, unpack(sch._A or {}))
          if r == nil then
            if dflt then
              if sch._T then
                addValToTable(conf, path, nil)
              end
            else
              if not sch._T or #path == 0 then path[#path+1] = '_ERROR' end
              addValToTable(errtab, path, 
                "ERR["..tostring(cnf).."]: must be "..e)
              res = nil
            end
          end
        end
      else -- no scheme
        local keyname = path[#path]
        if type(keyname) ~= 'string' or keyname:sub(1,1) ~= '_' then -- spec key
          if dflt then
            addValToTable(conf, path, nil)
          else
            addValToTable(errtab, path, 'No scheme for parameter')
            res = nil
          end
        end
      end
    else
      --addValToTable(errtab, path, {_E = 'No such parameter'})
      --res = nil
    end
  end
  return res, errtab
end

function getValByPath(t, path)
  t = t or {}
  local parentT
  for k, v in ipairs(path or {}) do
    parentT = t
    if type(t) == 'table' then 
      t = t[v]
    else
      t = nil
    end
    if not t then break end 
  end
  return t, parentT
end

local noManStr = "No manual for this parameter"

--- Search in current help tree
-- @param path path to node
-- @param scheme scheme table
-- @param conf configuration table
-- @param search searched string
function searchHelp(scheme, conf, path, search, locale)
   search = tostring(search):gsub("(%p)","%%%1")
   local hlp, man, hlpMatch, manMatch, nameMatch, hlpMark, manMark, nameMark, mark
   local helpTable={}
   local out={}
   for pth, sch, cnf in allNodes(path, scheme, conf, true, true) do
      if sch and #pth>0 then
         hlp = sch._H and getHelpString(help[sch._H], locale) or noHelpStr
         man = sch._H and getManual(sch, locale) or noManStr
         hlpMatch = hlp ~= noHelpStr and hlp:match(search)
         hlpMark = hlpMatch and "H" or "-"
         manMatch = man ~= noManStr and man:match(search)
         manMark = manMatch and "M" or "-"
         nameMatch =(tostring(pth[#pth])):match(search)
         nameMark = nameMatch and "N" or "-"
         mark = nameMark..hlpMark..manMark.." "
         if nameMatch or hlpMatch or manMatch
         then
            hlp = hlp or "nil"
            helpTable[hlp] = helpTable[hlp] or {}
            helpTable[hlp][#helpTable[hlp] + 1] = mark .. table.concat(pth,".")
         end
      end
   end
   for k,v in pairs(helpTable) do
      for _,p in ipairs(v) do
         out[#out + 1] = p
      end
      out[#out + 1] = k.."\n"
   end
   local _H = "searchMarking"
   _H = "searchNotFound"
   return table.concat( #out>0 and out or {getHelpString(help["searchNotFound"], locale)},'\n').."\n\n"..getHelpString(help["searchMarking"], locale)
end
 
--- Execute all apply functions(_Z)
-- @param path path to node
-- @param scheme scheme table
-- @param conf configuration table
function apply(path, scheme, conf, prevConf, first, sesNum)
  local errmes = {}
  for pth, sch, cnf in allNodes(path, scheme, conf, true, true) do
    if sch and type(sch._Z) == 'function' then
      local pcnf
      if first then
        if cnf == nil then pcnf = false end
      elseif prevConf then
        pcnf = getValByPath(prevConf, pth)
      end
      local ok, res, err = pcall(sch._Z, pathConcat(path, pth), sch, cnf, pcnf, nil, sesNum)
      if not ok then
        errmes[#errmes+1] = ser(pth)..": "..(res or 'apply error')
      elseif not res then 
        errmes[#errmes+1] = ser(pth)..": "..(err or 'apply error')
      end
    end
  end
  if #errmes == 0 then
    return true
  else
    return nil, table.concat(errmes, '\n')
  end
end

--- Execute function(_X)
-- @param path path to node
-- @param scheme scheme table
-- @param conf configuration table
function exec(path,scheme, conf, sesNum)
  if scheme and type(scheme._X) == 'function' then
    return toStr(scheme._X(path,scheme, conf, sesNum))
  end
  return "ERROR: no function defined"
end

--- Remove all needless value from configuration
-- @param scheme scheme table
-- @param conf configuration table
function stripConfig(absPath, scheme, conf)
  for path, sch, cnf in allNodes(absPath, scheme, conf) do
    if not sch then
      addValToTable(conf, path, nil)
    elseif sch._N then
      local need = sch._N(sch, cnf, path or {})
      for k, _ in sch:_pairs() do
        if need[k] ~= need[1] then cnf[k] = nil end
      end
    end
  end
end

--- Add default values to cofiguration
-- @param scheme scheme table
-- @param conf configuration table
function addDefault(absPath, scheme, conf, clone, opt, emptyVal)
  if clone then
    conf = deepcopy(conf) or {}
  else
    assert(conf, "addDefault: no configure table")
  end
  for path, sch, cnf in allNodes(absPath, scheme, conf) do
    if (cnf == nil and sch and
      (sch._D ~= nil or (emptyVal and sch._T and not sch._X)) and 
      (not opt or opt == 'onlylocal' and sch._LOCAL or 
      opt == 'onlyglobal' and not sch._LOCAL)
    ) then
      addValToTable(conf, path, sch._D or emptyVal)
    end
  end
  return conf
end

local function size (t)
  local n = 0
  for _ in pairs (t) do
    n = n + 1
  end
  return n
end

--- Compare two configurations
-- @param scheme scheme table
-- @param confA first configuration table
-- @param confB second configuration table
function confCmp(confA, confB)
  if confA == confB then return true end
  if type(confA) ~= 'table' or type(confB) ~= 'table' then return false end
  --if scheme then
  --  stripConfig(scheme, confA)
  --  stripConfig(scheme, confB)
  --  addDefault(path, scheme, confA)
  --  addDefault(path, scheme, confB)
  --end
  if size(confA) ~= size(confB) then return false end
  for k,v in pairs(confA) do
    if confB[k] == nil then return false end
    if not confCmp(v, confB[k]) then
      return false
    end
  end
  return true
end

noHelpStr = "<no help>"
---
function getHelpString(h, locale, noStr)
  if type(h) == 'string' then
    return h
  elseif type(h) == 'table' then
    if locale and h[locale] and not h[locale]:match'^%s*$' then
      return h[locale]
    elseif h[1] then
      return h[1]
    end
  end
  return noStr or noHelpStr
end

--[[
function printScheme(scheme, ...)
  local pars = {...}
  local set
  if #pars > 0 then
    set = {}
    for _, v in ipairs(pars) do set[v] = true end
  end
  local pt = {}
  for path, sch in allNodes(scheme) do
    local inx = #path+1
    if set then
      if set._D and sch._D then
        path[inx] = '_D'
        addValToTable(pt, path, sch._D)
      end
      if set._T and sch._T then
        path[inx] = '_T'
        addValToTable(pt, path, sch._T)
      end
      if set._H and sch._H then
        path[inx] = '_H'
        addValToTable(pt, path, getHelpString(help[sch._H]))
      end
      if set._V and sch._V then
        local r, e = sch._V(nil, unpack(sch._A or {}))
        path[inx] = '_V'
        addValToTable(pt, path, e)
      end
      if set._NAME_TYPE  and sch._NAME_TYPE  then
        local r, e = sch._NAME_TYPE (nil)
        path[inx] = '_NAM TYPE_'
        addValToTable(pt, path, e)
      end
    else
      if sch._T then
        local t = addValToTable(pt, path, sch._D or 'nil')
        if t ~= pt then return t end
      end
    end
    
    if sch._NAME then
      path[inx] = '_NAME'
      addValToTable(pt, path, printScheme(sch._NAME, ...))
    end
    if sch._LIST then
      path[inx] = '_LIST'
      addValToTable(pt, path, printScheme(sch._LIST, ...))
    end
  end
  return pt
end
--]]

function getChildScheme(path, sch, v)
  local path = clonePath(path, nil, v)
  if sch[v] then
    return sch[v](nil, path)
  elseif sch._LIST and type(v) == 'number' then
    return sch._LIST(nil, path)
  elseif sch._NAME and (sch._NAME_TYPE or isDefaultNameType)(v) then --TODO
    return sch._NAME(nil, path)
  end
end

--- Get help string for parameter in scheme
function getParameterHelp(path, sch, v, locale)
  if sch then
    path = clonePath(path, nil, v)
    if sch[v] and sch[v](nil, path) then
      return getHelpString(help[sch[v](nil, path)._H], locale)
    elseif sch._LIST and type(v) == 'number' then
      return getHelpString(help[sch._LIST(nil, path)._H], locale)
    elseif sch._NAME and (sch._NAME_TYPE or isDefaultNameType)(v) then --TODO
      return getHelpString(help[sch._NAME(nil, path)._H], locale)
    else
      return noHelpStr
    end
  else
    return noHelpStr
  end
end

--- Get manual string
function getManual(sch, locale)
  if sch._H and help[sch._H] then
     return getHelpString(help[sch._H].man, locale, noManStr.."("..tostring(sch._H)..")")
  else
     return noManStr.."("..tostring(sch._H)..")"
  end
end

do
  local function _() return nil end

  --- get subscheme
  function getSubScheme(path, sch)
    local key
    if type(path) == 'table' then
      key = path[#path]
    else
      key = path
      path = nil
    end
    if type(key) == "string" and key:sub(1,1) == '_'
    and (key ~= "_NAME" and key ~= "_LIST") then 
      return nil 
    end
    return (sch[key] or
      (type(key) == "number" and sch._LIST) or
      ((sch._NAME_TYPE or isDefaultNameType)(key) and sch._NAME) or _
    )(nil, path)
  end
end

--------------------------------------------------------------------------------
-- Validate functions
-- Must return the same or modified value(nil if error)

---
function anyVal(val)
  if type(val) == 'string' or type(val) == 'number' then
    return val
  else
    return nil, "string or number"
  end
end

function booleanVal(val)
  if type(val) == "boolean" then
    return val
  elseif type(val) == "string" then
    if val == ('true'):sub(1, #val) then
      return true
    elseif val == ('false'):sub(1, #val) then
      return false
    end
  end
  return nil, "boolean(true or false)"
end

---
function numberVal(val, min, max)
  if type(val) == "number" and val >= min and val <= max then
    return val
  end
  return nil, "number in range "..min..".."..max
end

---
function integerVal(val, min, max)
  if type(val) == "number" and val%1==0 and (not min or val >= min) 
  and (not max or val <= max) then
    return val
  end
  return nil, "integer in range "..(min or '-inf')..".."..(max or 'inf')
end

---
function stringVal(val, maxsize)
  if type(val) == "string"
  and (type(maxsize) ~= "number" or #val <= maxsize) then
    return val
  end
  return nil, "string"..(maxsize and 
    string.format('(maximum %d characters)', maxsize) or '')
end

---
function filenameVal(val)
  if stringVal(val) then
    return val
  end
  return nil, "file name"
end

---
function ipaddressVal(val)
  if type(val) == 'string' then
    local a1, a2, a3, a4 = 
    string.match(val,
    "^(%d%d?%d?)%.(%d%d?%d?)%.(%d%d?%d?)%.(%d%d?%d?)$")
    if a1 and
    tonumber(a1) <= 255 and
    tonumber(a2) <= 255 and
    tonumber(a3) <= 255 and
    tonumber(a4) <= 255 then
      return val
    end
  end
  return nil, "IP address in dot format, e.g '192.168.1.1'"
end

function ipprefixVal(val)
  if type(val) == 'string' then
    local a1, a2, a3, a4, m = 
    string.match(val,
    "^(%d%d?%d?)%.(%d%d?%d?)%.(%d%d?%d?)%.(%d%d?%d?)/(%d%d?)$")
    if a1 and
    tonumber(a1) <= 255 and
    tonumber(a2) <= 255 and
    tonumber(a3) <= 255 and
    tonumber(a4) <= 255 and
    tonumber(m)  <= 32  then
      return val
    end
  end
  return nil, "IP prefix, e.g '192.168.1.1/24'"
end

function ipprefixOrAddressVal(val)
   local ret1, err1 = ipprefixVal(val)
   local ret2, err2 = ipaddressVal(val)
   return (ret1 or ret2), not (ret1 or ret2) and (err1 or "").."\n"..(err2 or "")
end

function ipaddressAndPortVal(val)
  if type(val) == 'string' then
    local a1, a2, a3, a4, m = 
    string.match(val,
    "^(%d%d?%d?)%.(%d%d?%d?)%.(%d%d?%d?)%.(%d%d?%d?):(%d%d?%d?%d?%d?)$")
    if a1 and
    tonumber(a1) <= 255 and
    tonumber(a2) <= 255 and
    tonumber(a3) <= 255 and
    tonumber(a4) <= 255 and
    tonumber(m)  <= 0xFFFF  then
      return val
    end
  end
  return nil, "IP address and port, e.g '192.168.1.1:54321'"
end

---
function ipaddressOrAllVal(val)
  if val == '*' or ipaddressVal(val) then
    return val
  end
  return nil, 'IP address in dot format or "*" for any address'
end

---
function macaddressVal(val)
  if type(val) == 'string' then
    local a1, a2, a3, a4, a5, a6  =
    string.match(val, "^(%x%x)%:(%x%x):(%x%x)%:(%x%x):(%x%x)%:(%x%x)$")
    if a1 and
    tonumber('0x' .. a1) <= 255 and
    tonumber('0x' .. a2) <= 255 and
    tonumber('0x' .. a3) <= 255 and
    tonumber('0x' .. a4) <= 255 and
    tonumber('0x' .. a5) <= 255 and
    tonumber('0x' .. a6) <= 255
    then
      return val
    end
  end
  return nil, "MAC address in the format, e.g '00:09:56:0E:01:F3'"
end
---
function constantVal()
   return nil, "Value is CONSTANT and can`t be changed"
end
---
function choiceVal(val, ...)
  local var, all = {}, {}
  local maxLen = 0
  for _, v in ipairs{...} do
    local help
    if type(v) == 'table' then
      help = v[2]
      v = v[1]
    end
    if val == v then return val end
    all[#all+1] = {tostr(v), help}
    maxLen = math.max(maxLen, #tostr(v))
    if type(val) == 'string' and type(v) == 'string' then
      if val == v:sub(1,#val) then
        var[#var+1] = {v, help}
      end
    end
  end
  if #var == 1 then return var[1][1] end
  local prompt
  if #var == 0 then
    var = all; prompt = "variants:\n  "
  else
    prompt = 'completion:\n  '
    for k, v in ipairs(var) do
      var[k] = {tostr(v[1]), v[2]}
    end
  end
  for k, v in ipairs(var) do
    var[k] = v[1]..(' '):rep(maxLen + 1 - #v[1])..(v[2] or '')
  end
  local help = table.concat(var, '\n  ')
  return nil, prompt..help..'\n'
end

---
function areyousureVal(val)
  if val == 'yes' or val == 'no' then
    return val
  end
  return nil, '"yes" or "no"'
end

---
local function isSubset(t, list)
  local set = {}
  for _, v in ipairs(list) do set[v] = true end
  for _, v in ipairs(t) do
    if not set[v] then return false end
  end
  return true
end

---
function setVal(val, ...)
  local t = {...}
  if type(val) == 'string' then val = {val} end
  if type(val) == 'table' and isSubset(val, t) then return val end

  for k, v in ipairs(t) do
    t[k] = tostr(v)
  end
  local valList = table.concat(t, '\n  ')
  return nil, 'List of values in form {val1, val2, ..}, where val# are:\n  '..
  valList
end

---
function isString(val)
  if type(val) == 'string' and #val > 0 then
    return val
  else
    return nil, 'string'
  end
end

---
function isName(val)
  if type(val) == 'string' and #val > 0 and val:match'^%a' then
    return val
  else
    return nil, 'name(string where first char is letter)'
  end
end

---
function isNumber(val)
  if type(val) == 'number' then
    return val
  else
    return nil, 'number'
  end
end

---
function isStringOrNumber(val)
  if type(val) == 'string' and #val > 0  or type(val) == 'number' then
    return val
  else
    return nil, 'string or number'
  end
end

---
isDefaultNameType = isName
--------------------------------------------------------------------------------
-- Base types

--- Base boolean type definition
-- @class table
-- @name nilType
nilType = function()return nil end

--- Base boolean type definition
-- @class table
-- @name booleanType
booleanType = newScheme{
  _T = "boolean",
  _V = booleanVal,
  _H = "defaultBoolean",
}

--- Base string type definition
-- @class table
-- @name stringType
stringType = newScheme{
  _T = "string",
  _V = stringVal,
  _H = "defaultString",
}

--- Base integer type definition
-- @class table
-- @name integerType
integerType = newScheme{
  _T = "integer", 
  _V = integerVal,
}

--- Base tcp port type definition
-- @class table
-- @name tcpPortType
tcpPortType = newScheme{
  _T = "integer",
  _A = {0,0xffff}, 
  _V = integerVal, 
  _H = "defaultTcpPort",
}

--- Base IP address type definition
-- @class table
-- @name ipAddressType
ipAddressType = newScheme{
  _T = "ipaddr",
  _V = ipaddressVal,
  _H = "defaultIpaddress",
}  

--- Base "IP address or all" type definition
-- @class table
-- @name ipAddressOrAllType
ipAddressOrAllType = newScheme{
  _T = "string",
  _V = ipaddressOrAllVal,
  _H = "defaultIpaddressOrAll",
}

--- Base IP prefix type definition
-- @class table
-- @name ipPrefixType
ipPrefixType = newScheme{
  _T = "string",
  _V = ipprefixVal,
  _H = "defaultIpPrefix",
}

--- Base IP prefix/address type definition
-- @class table
-- @name ipPrefixTypeOrAddress
ipPrefixOrAddressType = newScheme{
  _T = "string",
  _V = ipprefixOrAddressVal,
  _H = "defaultIpPrefixOrAddress",
}

--- Base IP address and port type definition
-- @class table
-- @name ipAddressAndPortType
ipAddressAndPortType = newScheme{
  _T = "string",
  _V = ipaddressAndPortVal,
  _H = "defaultIpaddressAndPort",
}  

--- Base choice type definition
-- @class table
-- @name choiceType
choiceType = newScheme{
  _T = "choice",
  _V = choiceVal,
  _H = "defaultChoice",
}

--- Base constant type definition
-- @class table
-- @name constantType
constantType = stringType{
  _V = constantVal,
  _H = "defaultConstant",
}

--- Composite Types

IpAddressPool = newScheme{
  _H = "IpAddressPool",

  from = ipAddressType{_H = "IpAddressPoolFrom", _D = "172.16.0.2", },

  to = ipAddressType{_H = "IpAddressPoolTo", _D = "172.16.255.255", },
}
